# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Mappings on subsystems.
"""

from typing import Optional, Union, List
from qiskit import QiskitError

import numpy as np

from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics.systems import Subsystem, quantum_system_model
from .abstract_subsystem_operators import AbstractSubsystemOperator


class SubsystemMapping:
    r"""A linear mapping from a list of subsystems representing a tensor product space to another.

    This class represents a linear map :math:`A : V_1 \otimes \dots \otimes V_n \rightarrow W_1
    \otimes \dots \otimes W_m`, where :math:`A` is specified as a matrix, and the tensor factors of
    both the input and output spaces are given as lists of :class:`Subsystem` instances. The main
    usage is for mapping abstract operators or :class:`QuantumSystemModel` instances: the
    :meth:`.conjugate` method, or simply treating the mapping as ``Callable``, conjugates an
    operator or all operators within the :class:`QuantumSystemModel` by :math:`A`. As usual, for any
    subsystems in the ``in_subsystems`` of the mapping that the operator are not explicitly defined
    on, the operator is assumed to act as the identity.

    See the :ref:`How-to use advanced system modelling functionality <systems modelling userguide>`
    userguide entry for example usage of this class.
    """

    def __init__(
        self,
        matrix: np.ndarray,
        in_subsystems: Union[Subsystem, List[Subsystem]],
        out_subsystems: Optional[Union[Subsystem, List[Subsystem]]] = None,
    ):
        """Initialize.

        Args:
            matrix: The matrix form of the linear map.
            in_subsystems: The list of input subsystems.
            out_subsystems: The list of output subsystems.
        """
        if in_subsystems is None or in_subsystems == []:
            raise QiskitError("in_subsystems cannot be None or [] for SystemMapping.")

        if isinstance(in_subsystems, Subsystem):
            in_subsystems = [in_subsystems]

        if isinstance(out_subsystems, Subsystem):
            out_subsystems = [out_subsystems]
        elif out_subsystems is None:
            out_subsystems = in_subsystems

        self._in_subsystems = in_subsystems
        self._out_subsystems = out_subsystems

        in_dim = np.prod([x.dim for x in in_subsystems])
        out_dim = np.prod([x.dim for x in out_subsystems])

        if matrix.shape != (out_dim, in_dim):
            raise QiskitError("matrix.shape does not match input and output dimensions.")

        self._matrix = matrix
        self._matrix_adj = matrix.conj().transpose()

    @property
    def in_subsystems(self):
        """Subsystems for input to the mapping."""
        return self._in_subsystems

    @property
    def out_subsystems(self):
        """Subsystems for mapping output."""
        return self._out_subsystems

    @property
    def matrix(self):
        """Concrete matrix encoding the action of the mapping."""
        return self._matrix

    def conjugate(
        self, operator: Union[AbstractSubsystemOperator, "quantum_system_model.QuantumSystemModel"]
    ):
        r"""Conjugate a subsystem operator or model.

        Returns a subsystem operator representing :math:`A O A^\dagger`, where :math:`A` is the
        mapping matrix, and :math:`O` is the input operator. If applied to a
        :class:`QuantumSystemModel`, the mapping is applied to all operators in the model.

        Args:
            operator: The operator to be conjugated.
        Returns:
            Union[MappedOperator, QuantumSystemModel]: The conjugated operator or model.
        """

        if isinstance(operator, AbstractSubsystemOperator):
            return MappedOperator(operator, self)
        elif type(operator).__name__ == "QuantumSystemModel":
            return operator._map_model(lambda x: MappedOperator(x, self))

        raise QiskitError(
            f"Input of type {type(operator)} not recognized by SubsystemMapping.conjugate."
        )

    def __call__(
        self, operator: Union[AbstractSubsystemOperator, "quantum_system_model.QuantumSystemModel"]
    ):
        """Apply the conjugation function."""
        return self.conjugate(operator)


class MappedOperator(AbstractSubsystemOperator):
    """An operator mapped by a SystemMapping."""

    def __init__(self, operator, system_mapping):
        """Initialize."""

        self._operator = operator
        self._system_mapping = system_mapping

        # validate that none of the out_subsystems of system_mapping are in the operator definition
        if any(s in system_mapping.out_subsystems for s in operator.subsystems):
            raise QiskitError("Output subsystem found in input operator subsystems.")

        unmapped_subsystems = []
        for s in operator.subsystems:
            if s not in system_mapping.in_subsystems:
                unmapped_subsystems.append(s)

        self._unmapped_subsystems = unmapped_subsystems
        super().__init__(system_mapping.out_subsystems + unmapped_subsystems)

    @property
    def operator(self):
        """The operator being mapped."""
        return self._operator

    @property
    def system_mapping(self):
        """The system mapping being applied to the operator."""
        return self._system_mapping

    @property
    def unmapped_subsystems(self):
        """Subsystems of underlying operator unaffected by mapping."""
        return self._unmapped_subsystems

    def base_matrix(self):
        mat = self._operator.matrix(self.system_mapping.in_subsystems + self.unmapped_subsystems)
        A = self.system_mapping.matrix
        if len(self.unmapped_subsystems) > 0:
            unmapped_dim = np.prod([x.dim for x in self.unmapped_subsystems])
            A = unp.kron(np.eye(unmapped_dim, dtype=complex), A)
        return A @ mat @ A.conj().transpose()

    def __str__(self):
        return f"""MappedOperator({self._operator}, {self.system_mapping.in_subsystems} ->
                {self.system_mapping.out_subsystems})"""



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Module for rotating frame handling classes.
"""

from typing import Union, List, Optional
import numpy as np
from scipy.sparse import issparse

from qiskit import QiskitError
from qiskit.quantum_info.operators import Operator
from qiskit.quantum_info.operators.predicates import is_hermitian_matrix
from qiskit_dynamics.arraylias import ArrayLike
from qiskit_dynamics.arraylias import DYNAMICS_NUMPY_ALIAS
from qiskit_dynamics.arraylias import DYNAMICS_NUMPY as unp
from qiskit_dynamics.arraylias.alias import _numpy_multi_dispatch

try:
    import jax.numpy as jnp
except ImportError:
    pass


class RotatingFrame:
    r"""Class for representing a rotation frame transformation.

    This class provides functionality for transforming various objects into or out-of a rotating
    frame specified by an anti-Hermitian operator :math:`F = -iH`. For example:

        * Bringing a "state" into/out of the frame: :math:`t, y \mapsto e^{\mp tF}y`
        * Bringing an "operator" into/out of the frame: :math:`t, A \mapsto e^{\mp tF}Ae^{\pm tF}`
        * Bringing a generator for a LMDE into/out of the frame:
          :math:`t, G \mapsto e^{\mp tF}Ge^{\pm tF} - F`

    This class also contains functions for bringing states/operators into/out of the basis in which
    :math:`F` is diagonalized, which we refer to as the "frame basis". All previously mentioned
    functions also include optional arguments specifying whether the input/output are meant to be in
    the frame basis.

    .. note::
        :class:`~qiskit_dynamics.models.RotatingFrame` can be instantiated
        with a 1d array, which is understood to correspond to the diagonal entries
        of a diagonal :math:`H` or :math:`F = -i H`.

    """

    def __init__(
        self,
        frame_operator: Union[ArrayLike, Operator, None],
        atol: float = 1e-10,
        rtol: float = 1e-10,
    ):
        """Initialize with a frame operator.

        Args:
            frame_operator: The frame operator, which must be either Hermitian or anti-Hermitian.
            atol: Absolute tolerance when verifying that the ``frame_operator`` is Hermitian or
                anti-Hermitian.
            rtol: Relative tolerance when verifying that the ``frame_operator`` is Hermitian or
                anti-Hermitian.
        """
        if isinstance(frame_operator, RotatingFrame):
            frame_operator = frame_operator.frame_operator

        self._frame_operator = frame_operator
        if frame_operator is not None:
            frame_operator = unp.asarray(frame_operator)

        if frame_operator is None:
            self._dim = None
            self._frame_diag = None
            self._frame_basis = None
            self._frame_basis_adjoint = None
        # if frame_operator is a 1d array, assume already diagonalized
        elif frame_operator.ndim == 1:
            # verify Hermitian or anti-Hermitian
            # if Hermitian convert to anti-Hermitian
            frame_operator = _enforce_anti_herm(frame_operator, atol=atol, rtol=rtol)

            self._frame_diag = frame_operator
            self._frame_basis = None
            self._frame_basis_adjoint = None
            self._dim = len(self._frame_diag)
        # if not, diagonalize it
        else:
            # verify Hermitian or anti-Hermitian
            # if Hermitian convert to anti-Hermitian
            frame_operator = _enforce_anti_herm(frame_operator, atol=atol, rtol=rtol)

            # diagonalize with eigh, utilizing assumption of anti-hermiticity
            frame_diag, frame_basis = unp.linalg.eigh(1j * frame_operator)

            self._frame_diag = -1j * frame_diag
            self._frame_basis = frame_basis
            self._frame_basis_adjoint = frame_basis.conj().transpose()
            self._dim = len(self._frame_diag)

        # lazily evaluate change-of-basis matrices for vectorized operators.
        self._vectorized_frame_basis = None
        self._vectorized_frame_basis_adjoint = None

    @property
    def dim(self) -> int:
        """The dimension of the frame."""
        return self._dim

    @property
    def frame_operator(self) -> ArrayLike:
        """The original frame operator."""
        return self._frame_operator

    @property
    def frame_diag(self) -> ArrayLike:
        """The diagonal of the frame operator."""
        return self._frame_diag

    @property
    def frame_basis(self) -> ArrayLike:
        """The array containing diagonalizing unitary."""
        return self._frame_basis

    @property
    def frame_basis_adjoint(self) -> ArrayLike:
        """The adjoint of the diagonalizing unitary."""
        return self._frame_basis_adjoint

    def state_into_frame_basis(self, y: ArrayLike) -> ArrayLike:
        r"""Take a state into the frame basis, i.e. return ``self.frame_basis_adjoint @ y``.

        Args:
            y: The state.

        Returns:
            ArrayLike: The state in the frame basis.
        """
        y = unp.asarray(y)
        if self.frame_basis_adjoint is None:
            return y

        return self.frame_basis_adjoint @ y

    def state_out_of_frame_basis(self, y: ArrayLike) -> ArrayLike:
        r"""Take a state out of the frame basis, i.e. ``return self.frame_basis @ y``.

        Args:
            y: The state.

        Returns:
            ArrayLike: The state in the frame basis.
        """
        y = unp.asarray(y)
        if self.frame_basis is None:
            return y

        return self.frame_basis @ y

    def operator_into_frame_basis(
        self,
        op: Union[Operator, List[Operator], ArrayLike, None],
        convert_type: bool = True,
    ) -> ArrayLike:
        r"""Take an operator into the frame basis, i.e. return
        ``self.frame_basis_adjoint @ A @ self.frame_basis``

        Args:
            op: The operator or array of operators.
            convert_type: Whether or not to initially convert ``op`` into an expected type. Should
                          only be set to ``False`` in situations in which it is gauranteed that
                          ``op`` is a handled input type.

        Returns:
            ArrayLike: The operator in the frame basis.
        """
        if convert_type:
            op = unp.asarray(op)

        if self.frame_basis is None or op is None:
            return op

        if isinstance(op, list):
            return [self.operator_into_frame_basis(x, convert_type=False) for x in op]

        return unp.rmatmul(unp.matmul(op, self.frame_basis), self.frame_basis_adjoint)

    def operator_out_of_frame_basis(
        self,
        op: Union[Operator, List[Operator], ArrayLike, None],
        convert_type: bool = True,
    ) -> ArrayLike:
        r"""Take an operator out of the frame basis, i.e. return
        ``self.frame_basis @ op @ self.frame_basis_adjoint``.

        Args:
            op: The operator or array of operators.
            convert_type: Whether or not to initially convert ``op`` into an expected type. Should
                          only be set to ``False`` in situations in which it is gauranteed that
                          that ``op`` is a handled input type.

        Returns:
            ArrayLike: The operator in the frame basis.
        """
        if convert_type:
            op = unp.asarray(op)

        if self.frame_basis is None or op is None:
            return op

        if isinstance(op, list) or (isinstance(op, np.ndarray) and issparse(op[0])):
            return [self.operator_out_of_frame_basis(x, convert_type=False) for x in op]

        return unp.rmatmul(unp.matmul(op, self.frame_basis_adjoint), self.frame_basis)

    def state_into_frame(
        self,
        t: float,
        y: ArrayLike,
        y_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
    ) -> ArrayLike:
        """Take a state into the rotating frame, i.e. return ``exp(-tF) @ y``.

        Args:
            t: The time.
            y: The state.
            y_in_frame_basis: Whether or not the array y is already in the basis in which the frame
                is diagonal.
            return_in_frame_basis: Whether or not to return the result in the frame basis.

        Returns:
            ArrayLike: The state in the rotating frame.
        """
        y = unp.asarray(y)
        if self._frame_operator is None:
            return y

        out = y

        # if not in frame basis convert it
        if not y_in_frame_basis:
            out = self.state_into_frame_basis(out)

        # go into the frame using fast diagonal matrix multiplication
        out = (unp.exp(self.frame_diag * (-t)) * out.transpose()).transpose()  # = e^{tF}out

        # if output is requested to not be in the frame basis, convert it
        if not return_in_frame_basis:
            out = self.state_out_of_frame_basis(out)

        return out

    def state_out_of_frame(
        self,
        t: float,
        y: ArrayLike,
        y_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
    ) -> ArrayLike:
        r"""Take a state out of the rotating frame, i.e. return ``exp(tF) @ y``.

        Calls ``self.state_into_frame`` with time reversed.

        Args:
            t: The time.
            y: The state..
            y_in_frame_basis: Whether or not the array y is already in the basis in which the frame
                is diagonal.
            return_in_frame_basis: Whether or not to return the result in the frame basis.

        Returns:
            ArrayLike: The state out of the rotating frame.
        """
        return self.state_into_frame(-t, y, y_in_frame_basis, return_in_frame_basis)

    def _conjugate_and_add(
        self,
        t: float,
        operator: ArrayLike,
        op_to_add_in_fb: Optional[ArrayLike] = None,
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
        vectorized_operators: Optional[bool] = False,
    ) -> ArrayLike:
        r"""General helper function for computing :math:`\exp(-tF)G\exp(tF) + B`.

        Note: :math:`B` is added in the frame basis before any potential final change
        out of the frame basis.

        Note that there are two conventions for passing multiple operators at the same time. For
        evaluation with ``vectorized_operators=False``, these operators should be passed as ``(k,
        dim, dim)`` array objects, with the :math:`i^{th}` operator being stored as the ``[i,:,:]``
        entry. For ``vectorized_operators = True``, these (vectorized) operators should be passed as
        a ``(dim**2, k)`` array, with the :math:`i^{th}` vectorized operator stored as the ``[:,i]``
        entry.

        Args:
            t: The time.
            operator: The operator :math:`G`.
            op_to_add_in_fb: The additional operator :math:`B`.
            operator_in_fame_basis: Whether ``operator`` is already in the basis in which the frame
                operator is diagonal.
            vectorized_operators: Whether ``operator`` is passed as a vectorized, ``(dim**2,)``
                array, rather than a ``(dim,dim)`` array.

        Returns:
            Array of the newly conjugated operator.
        """
        operator = unp.asarray(operator)
        if op_to_add_in_fb is not None:
            op_to_add_in_fb = unp.asarray(op_to_add_in_fb)
        if vectorized_operators:
            # If passing vectorized operator, undo vectorization temporarily
            if self._frame_operator is None:
                if op_to_add_in_fb is None:
                    return operator
                else:
                    return operator + op_to_add_in_fb
            if len(operator.shape) == 2:
                operator = operator.T
            operator = operator.reshape(operator.shape[:-1] + (self.dim, self.dim), order="F")

        if self._frame_operator is None:
            if op_to_add_in_fb is None:
                return operator
            else:
                op_to_add_in_fb = DYNAMICS_NUMPY_ALIAS(like=operator).asarray(op_to_add_in_fb)

                return operator + op_to_add_in_fb

        out = operator

        # if not in frame basis convert it
        if not operator_in_frame_basis:
            out = self.operator_into_frame_basis(out)

        # get frame transformation matrix in diagonal basis
        # assumption that F is anti-Hermitian implies conjugation of
        # diagonal gives inversion
        exp_freq = unp.exp(self.frame_diag * t)
        frame_mat = exp_freq.conj().reshape(self.dim, 1) * exp_freq

        out = _numpy_multi_dispatch(out, frame_mat, path="multiply")

        if op_to_add_in_fb is not None:
            op_to_add_in_fb = DYNAMICS_NUMPY_ALIAS(like=out).asarray(op_to_add_in_fb)

            out = out + op_to_add_in_fb

        # if output is requested to not be in the frame basis, convert it
        if not return_in_frame_basis:
            out = self.operator_out_of_frame_basis(out)

        if vectorized_operators:
            # If a vectorized output is required, reshape correctly
            out = out.reshape(out.shape[:-2] + (self.dim**2,), order="F")
            if len(out.shape) == 2:
                out = out.T

        return out

    def operator_into_frame(
        self,
        t: float,
        operator: Union[Operator, ArrayLike],
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
        vectorized_operators: Optional[bool] = False,
    ) -> ArrayLike:
        r"""Bring an operator into the frame, i.e. return
        ``exp(-tF) @ operator @ exp(tF)``.

        The default implementation is to use ``self._conjugate_and_add``.

        Args:
            t: The time.
            operator: An array of appropriate size.
            operator_in_frame_basis: Whether or not the operator is already in the basis in which
                the frame is diagonal.
            return_in_frame_basis: Whether or not to return the result in the frame basis.
            vectorized_operators: Whether ``operator`` is passed as a vectorized,
                ``(dim**2,)`` array, rather than a ``(dim,dim)`` array.

        Returns:
            ArrayLike: The operator in the rotating frame.
        """
        return self._conjugate_and_add(
            t,
            operator,
            operator_in_frame_basis=operator_in_frame_basis,
            return_in_frame_basis=return_in_frame_basis,
            vectorized_operators=vectorized_operators,
        )

    def operator_out_of_frame(
        self,
        t: float,
        operator: Union[Operator, ArrayLike],
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
        vectorized_operators: Optional[bool] = False,
    ) -> ArrayLike:
        r"""Bring an operator into the rotating frame, i.e. return
        ``exp(tF) @ operator @ exp(-tF)``.

        The default implmentation is to use `self.operator_into_frame`.

        Args:
            t: Time.
            operator: An array of appropriate size.
            operator_in_frame_basis: Whether or not the operator is already in the basis in which
                the frame is diagonal.
            return_in_frame_basis: Whether or not to return the result in the frame basis.
            vectorized_operators: Whether ``operator`` is passed as a vectorized, ``(dim**2,)``
                array, rather than a ``(dim,dim)`` array.

        Returns:
            ArrayLike: The operator out of the rotating frame.
        """
        return self.operator_into_frame(
            -t,
            operator,
            operator_in_frame_basis=operator_in_frame_basis,
            return_in_frame_basis=return_in_frame_basis,
            vectorized_operators=vectorized_operators,
        )

    def generator_into_frame(
        self,
        t: float,
        operator: Union[Operator, ArrayLike],
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
        vectorized_operators: Optional[bool] = False,
    ) -> ArrayLike:
        r"""Take an generator into the rotating frame, i.e. return
        ``exp(-tF) @ operator @ exp(tF) - F``.

        The default implementation is to use :meth:`_conjugate_and_add`.

        Args:
            t: Time.
            operator: Generator (array of appropriate size).
            operator_in_frame_basis: Whether or not the generator is already in the basis in which
                the frame is diagonal.
            return_in_frame_basis: Whether or not to return the result in the frame basis.
            vectorized_operators: Whether ``operator`` is passed as a vectorized, ``(dim**2,)``
                array, rather than a ``(dim,dim)`` array.

        Returns:
            ArrayLike: The generator in the rotating frame.
        """
        if self.frame_operator is None:
            return unp.asarray(operator)
        else:
            # conjugate and subtract the frame diagonal
            return self._conjugate_and_add(
                t,
                operator,
                op_to_add_in_fb=-unp.diag(self.frame_diag),
                operator_in_frame_basis=operator_in_frame_basis,
                return_in_frame_basis=return_in_frame_basis,
                vectorized_operators=vectorized_operators,
            )

    def generator_out_of_frame(
        self,
        t: float,
        operator: Union[Operator, ArrayLike],
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
    ) -> ArrayLike:
        r"""Take an operator out of the frame using the generator transformaton rule, i.e. return
        ``exp(tF) @ operator @ exp(-tF) + F``.

        The default implementation is to use ``self._conjugate_and_add``.

        Args:
            t: The time
            operator: Generator (array of appropriate size).
            operator_in_frame_basis: Whether or not the operator is already in the basis in which
                the frame is diagonal.
            return_in_frame_basis: Whether or not to return the result in the frame basis.

        Returns:
            ArrayLike: The generator out of the rotating frame.
        """
        if self.frame_operator is None:
            return unp.asarray(operator)
        else:
            # conjugate and add the frame diagonal
            return self._conjugate_and_add(
                -t,
                operator,
                op_to_add_in_fb=unp.diag(self.frame_diag),
                operator_in_frame_basis=operator_in_frame_basis,
                return_in_frame_basis=return_in_frame_basis,
            )

    @property
    def vectorized_frame_basis(self):
        """Lazily evaluated operator for mapping vectorized operators into the frame basis."""

        if self.frame_basis is None:
            return None

        if self._vectorized_frame_basis is None:
            self._vectorized_frame_basis = unp.kron(self.frame_basis.conj(), self.frame_basis)
            self._vectorized_frame_basis_adjoint = self._vectorized_frame_basis.conj().transpose()

        return self._vectorized_frame_basis

    @property
    def vectorized_frame_basis_adjoint(self):
        """Lazily evaluated operator for mapping vectorized operators out of the frame basis."""

        if self.frame_basis is None:
            return None

        if self._vectorized_frame_basis_adjoint is None:
            # trigger lazy evaluation of vectorized_frame_basis
            # pylint: disable=pointless-statement
            self.vectorized_frame_basis

        return self._vectorized_frame_basis_adjoint

    def vectorized_map_into_frame(
        self,
        time: float,
        op: ArrayLike,
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
    ) -> ArrayLike:
        r"""Given an operator ``op`` of dimension ``dim**2`` assumed to represent vectorized linear
        map in column stacking convention, returns:

        .. math::
            ((e^{tF})^T \otimes e^{-tF}) \times op \times ((e^{-tF})^T \otimes e^{tF}).

        Utilizes element-wise multiplication :math:`op \to \Delta\otimes\bar{\Delta} \odot op`,
        where :math:`\Delta_{ij}=\exp((-d_i+d_j)t)` is the frame difference matrix, as well as
        caches array :math:`\bar{C}\otimes C`, where ``C = self.frame_basis`` for future use.

        Args:
            time: The time :math:`t`.
            op: The ``(dim**2,dim**2)`` array.
            operator_in_frame_basis: Whether the operator is in the frame basis.
            return_in_frame_basis: Whether the operator should be returned in the frame basis.

        Returns:
            ArrayLike: ``op`` in the frame.
        """
        if self.frame_diag is not None:
            # Put the vectorized operator into the frame basis
            if not operator_in_frame_basis and self.frame_basis is not None:
                op = self.vectorized_frame_basis_adjoint @ (op @ self.vectorized_frame_basis)

            expvals = np.exp(self.frame_diag * time)  # = e^{td_i} = e^{it*Im(d_i)}
            # = kron(e^{-it*Im(d_i)},e^{it*Im(d_i)}), but ~3x faster
            temp_outer = (expvals.conj().reshape(self.dim, 1) * expvals).flatten()
            delta_bar_otimes_delta = np.outer(
                temp_outer.conj(), temp_outer
            )  # = kron(delta.conj(),delta) but >3x faster
            if issparse(op):
                op = op.multiply(delta_bar_otimes_delta)
            else:
                op = delta_bar_otimes_delta * op  # hadamard product

            if not return_in_frame_basis and self.frame_basis is not None:
                op = self.vectorized_frame_basis @ (op @ self.vectorized_frame_basis_adjoint)

        return op


def _enforce_anti_herm(
    mat: ArrayLike, atol: Optional[float] = 1e-10, rtol: Optional[float] = 1e-10
) -> ArrayLike:
    r"""Given ``mat``, the logic of this function is:
        - if ``mat`` is hermitian, return ``-1j * mat``
        - if ``mat`` is anti-hermitian, return ``mat``
        - otherwise:
            - if ``mat.backend == 'jax'`` return ``jnp.inf * mat``
            - otherwise raise an error

    The main purpose of this function is to hide the pecularities of implementing the above logic in
    a compileable way in ``jax``.

    Args:
        mat: The array to check.
        atol: The absolute tolerance.
        rtol: The relative tolerance.

    Returns:
        ArrayLike: Anti-hermitian version of ``mat``, if applicable.

    Raises:
        ImportError: If the backend is jax and jax is not installed.
        QiskitError: If ``mat`` is not Hermitian or anti-Hermitian.
    """
    mat = unp.asarray(mat)

    try:
        from jax.dtypes import canonicalize_dtype
        from jax.lax import cond, convert_element_type

        if isinstance(unp.asarray(mat), jnp.ndarray):
            complex_dtype = canonicalize_dtype(jnp.complex128)

            if mat.ndim == 1:
                # this function checks if pure imaginary. If yes it returns the
                # array, otherwise it multiplies it by jnp.nan to raise an error
                # Note: pathways in conditionals in jax cannot raise Exceptions
                def anti_herm_conditional(b):
                    aherm_pred = jnp.allclose(b, -b.conj(), atol=atol, rtol=rtol)
                    return convert_element_type(
                        cond(aherm_pred, lambda A: A, lambda A: jnp.nan * A, b), complex_dtype
                    )

                # Check if it is purely real, if not apply anti_herm_conditional
                herm_pred = jnp.allclose(mat, mat.conj(), atol=atol, rtol=rtol)
                return unp.asarray(cond(herm_pred, lambda A: -1j * A, anti_herm_conditional, mat))
            else:
                # this function checks if anti-hermitian, if yes returns the array,
                # otherwise it multiplies it by jnp.nan
                def anti_herm_conditional(b):
                    aherm_pred = jnp.allclose(b, -b.conj().transpose(), atol=atol, rtol=rtol)
                    return convert_element_type(
                        cond(aherm_pred, lambda A: A, lambda A: jnp.nan * A, b), complex_dtype
                    )

                # the following lines check if a is hermitian, otherwise it feeds
                # it into the anti_herm_conditional
                herm_pred = jnp.allclose(mat, mat.conj().transpose(), atol=atol, rtol=rtol)
                return unp.asarray(cond(herm_pred, lambda A: -1j * A, anti_herm_conditional, mat))
    except (ImportError, NameError):
        pass

    if mat.ndim == 1:
        if np.allclose(mat, mat.conj(), atol=atol, rtol=rtol):
            return -1j * mat
        elif np.allclose(mat, -mat.conj(), atol=atol, rtol=rtol):
            return mat
    else:
        if is_hermitian_matrix(mat, rtol=rtol, atol=atol):
            return -1j * mat
        elif is_hermitian_matrix(1j * mat, rtol=rtol, atol=atol):
            return mat

    # raise error if execution has made it this far
    raise QiskitError("""frame_operator must be either a Hermitian or anti-Hermitian matrix.""")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Concrete subsystem operators.
"""

import numpy as np

from .abstract_subsystem_operators import AbstractSubsystemOperator


class SubsystemOperator(AbstractSubsystemOperator):
    """A concrete operator specified in terms of a matrix."""

    def __init__(self, matrix, subsystems, str_label=None):
        """Initialize.

        Args:
            matrix: The matrix of the operator.
            subsystems: The ordered subsystems representing the tensor factor system the matrix is
                specified on.
        """
        if matrix.shape[0] != np.prod([s.dim for s in subsystems]):
            raise ValueError("Subsystem dimensions don't match matrix shape.")

        self._matrix = matrix
        self._str_label = str_label
        super().__init__(subsystems)

    def base_matrix(self):
        return self._matrix

    def __str__(self):
        str_label = self._str_label or "SubsystemOperator"

        subsystem_str = str(self.subsystems[0])
        for s in self.subsystems[1:]:
            subsystem_str += f", {s}"
        return f"{str_label}({subsystem_str})"


class A(AbstractSubsystemOperator):
    r"""Annihilation operator.

    Defined as the matrix with non-zero entries :math:`0, 1, \sqrt{2}, ..., \sqrt{n - 1}` in the
    first off-diagonal, where :math:`n` is the dimension of the subsystem being acted on.
    """

    def base_matrix(self):
        return np.diag(np.sqrt(np.arange(1, self.subsystems[0].dim, dtype=complex)), 1)

    def __str__(self):
        return f"a({self.subsystems[0]})"


class Adag(AbstractSubsystemOperator):
    r"""Creation operator.

    Defined as the matrix with non-zero entries :math:`0, 1, \sqrt{2}, ..., \sqrt{n - 1}` in the
    first lower off-diagonal, where :math:`n` is the dimension of the subsystem being acted on.
    """

    def base_matrix(self):
        return np.diag(np.sqrt(np.arange(1, self.subsystems[0].dim, dtype=complex)), -1)

    def __str__(self):
        return f"adag({self.subsystems[0]})"


class N(AbstractSubsystemOperator):
    """The number operator.

    Defined as the diagonal matrix with with entries ``[0, ..., dim - 1]``, where ``dim`` is the
    dimension of the :class:`Subsystem` the operator is defined on.
    """

    def base_matrix(self):
        return np.diag(np.arange(self.subsystems[0].dim, dtype=complex))

    def __str__(self):
        return f"N({self.subsystems[0]})"


class I(AbstractSubsystemOperator):  # noqa: E742
    """The identity operator."""

    def base_matrix(self):
        return np.eye(self.subsystems[0].dim, dtype=complex)

    def __str__(self):
        return f"I({self.subsystems[0]})"


class X(AbstractSubsystemOperator):
    """X operator.

    The standard Pauli :math:`X` operator, generalized to ``A + Adag`` for higher dimensions.
    """

    def base_matrix(self):
        return A(self.subsystems).base_matrix() + Adag(self.subsystems).base_matrix()

    def __str__(self):
        return f"X({self.subsystems[0]})"


class Y(AbstractSubsystemOperator):
    """Y operator.

    The standard Pauli :math:`Y` operator, generalized to ``-1j * (A - Adag)`` for higher
    dimensions.
    """

    def base_matrix(self):
        return -1j * (A(self.subsystems).base_matrix() + (-1 * Adag(self.subsystems).base_matrix()))

    def __str__(self):
        return f"Y({self.subsystems[0]})"


class Z(AbstractSubsystemOperator):
    """Z operator.

    The standard Pauli :math:`Z` operator, generalized to ``I - 2 * N`` for higher dimensions.
    """

    def base_matrix(self):
        return I(self.subsystems).base_matrix() - 2 * N(self.subsystems).base_matrix()

    def __str__(self):
        return f"Z({self.subsystems[0]})"



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Hamiltonian models module.
"""

from typing import Union, List, Optional
import numpy as np
from scipy.sparse import issparse
from scipy.sparse.linalg import norm as spnorm

from qiskit import QiskitError

from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics import DYNAMICS_NUMPY_ALIAS as numpy_alias
from qiskit_dynamics.arraylias.alias import ArrayLike, _isArrayLike
from qiskit_dynamics.signals import Signal, SignalList
from .generator_model import GeneratorModel
from .rotating_frame import RotatingFrame


class HamiltonianModel(GeneratorModel):
    r"""A model of a Hamiltonian for the Schrodinger equation.

    This class represents a Hamiltonian as a time-dependent decomposition the form:

    .. math::
        H(t) = H_d + \sum_j s_j(t) H_j,

    where :math:`H_j` are Hermitian operators, :math:`H_d` is the static component, and the
    :math:`s_j(t)` are either :class:`~qiskit_dynamics.signals.Signal` objects or numerical
    constants. Constructing a :class:`~qiskit_dynamics.models.HamiltonianModel` requires specifying
    the above decomposition, e.g.:

    .. code-block:: python

        hamiltonian = HamiltonianModel(
            static_operator=static_operator,
            operators=operators,
            signals=signals
        )

    This class inherits most functionality from :class:`GeneratorModel`, with the following
    modifications:

        * The operators :math:`H_d` and :math:`H_j` are assumed and verified to be Hermitian.
        * Rotating frames are dealt with assuming the structure of the Schrodinger equation. I.e.
          Evaluating the Hamiltonian :math:`H(t)` in a frame :math:`F = -iH_0`, evaluates the
          expression :math:`e^{-tF}H(t)e^{tF} - H_0`.
    """

    def __init__(
        self,
        static_operator: Optional[ArrayLike] = None,
        operators: Optional[ArrayLike] = None,
        signals: Optional[Union[SignalList, List[Signal]]] = None,
        rotating_frame: Optional[Union[ArrayLike, RotatingFrame]] = None,
        in_frame_basis: bool = False,
        array_library: Optional[str] = None,
        validate: bool = True,
    ):
        """Initialize, ensuring that the operators are Hermitian.

        Args:
            static_operator: Time-independent term in the Hamiltonian.
            operators: List of Operator objects.
            signals: List of coefficients :math:`s_i(t)`. Not required at instantiation, but
                necessary for evaluation of the model.
            rotating_frame: Rotating frame operator. If specified with a 1d array, it is interpreted
                as the diagonal of a diagonal matrix. Assumed to store the antihermitian matrix
                F = -iH.
            in_frame_basis: Whether to represent the model in the basis in which the rotating
                frame operator is diagonalized.
            array_library: Array library with which to represent the operators in the model, and to
                evaluate the model. See the list of supported array libraries in the
                :mod:`.arraylias` submodule API documentation. If ``None``, the arrays will be
                handled by general dispatching rules.
            validate: If ``True`` check input operators are Hermitian. Note that this is
                incompatible with JAX transformations.

        Raises:
            QiskitError: if operators are not Hermitian
        """

        # prepare and validate operators
        if static_operator is not None:
            if validate and not is_hermitian(static_operator):
                raise QiskitError("""HamiltonianModel static_operator must be Hermitian.""")
            static_operator = -1j * numpy_alias(like=array_library).asarray(static_operator)

        if operators is not None:
            if validate and any(not is_hermitian(op) for op in operators):
                raise QiskitError("""HamiltonianModel operators must be Hermitian.""")

            if array_library == "scipy_sparse" or (
                array_library is None and "scipy_sparse" in numpy_alias.infer_libs(operators)
            ):
                operators = [-1j * numpy_alias(like=array_library).asarray(op) for op in operators]
            else:
                operators = -1j * numpy_alias(like=array_library).asarray(operators)

        super().__init__(
            static_operator=static_operator,
            operators=operators,
            signals=signals,
            rotating_frame=rotating_frame,
            in_frame_basis=in_frame_basis,
            array_library=array_library,
        )

    @property
    def static_operator(self) -> Union[ArrayLike, None]:
        """The static operator."""
        if self._operator_collection.static_operator is None:
            return None

        if self.in_frame_basis:
            return self._operator_collection.static_operator
        return 1j * self.rotating_frame.operator_out_of_frame_basis(
            self._operator_collection.static_operator
        )

    @property
    def operators(self) -> Union[ArrayLike, None]:
        """The operators in the model."""
        if self._operator_collection.operators is None:
            return None

        if self.in_frame_basis:
            ops = self._operator_collection.operators
        else:
            ops = self.rotating_frame.operator_out_of_frame_basis(
                self._operator_collection.operators
            )

        if isinstance(ops, list):
            return [1j * op for op in ops]

        return 1j * ops


def is_hermitian(operator: ArrayLike, tol: Optional[float] = 1e-10) -> bool:
    """Validate that an operator is Hermitian.

    Args:
        operator: A 2d array representing a single operator.
        tol: Tolerance for checking zeros.

    Returns:
        bool: Whether or not the operator is Hermitian to within tolerance.

    Raises:
        QiskitError: If an unexpeted type is received.
    """
    operator = unp.asarray(operator)

    if issparse(operator):
        return spnorm(operator - operator.conj().transpose()) < tol
    elif type(operator).__name__ == "BCOO":
        # fall back on array case for BCOO
        return is_hermitian(operator.todense())
    elif _isArrayLike(operator):
        adj = None
        adj = unp.transpose(unp.conjugate(operator))
        return np.linalg.norm(adj - operator) < tol

    raise QiskitError("is_hermitian got an unexpected type.")



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
==========================================
Qiskit Dynamics (:mod:`qiskit_dynamics`)
==========================================

.. currentmodule:: qiskit_dynamics

Qiskit extension module for simulating quantum dynamics.
"""
from .version import __version__

from .arraylias.alias import (
    DYNAMICS_NUMPY_ALIAS,
    DYNAMICS_SCIPY_ALIAS,
    DYNAMICS_NUMPY,
    DYNAMICS_SCIPY,
    ArrayLike,
)

from .models.rotating_frame import RotatingFrame

from .signals.signals import Signal, DiscreteSignal

from .solvers.solver_functions import solve_ode, solve_lmde
from .solvers.solver_classes import Solver
from .solvers.perturbative_solvers.dyson_solver import DysonSolver
from .solvers.perturbative_solvers.magnus_solver import MagnusSolver
from .backend.dynamics_backend import DynamicsBackend

from . import models
from . import signals
from . import pulse
from . import backend



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Registering transpose.
"""


def register_transpose(alias):
    """Register linear functions for each array library."""

    try:
        from jax.experimental.sparse import bcoo_transpose

        @alias.register_function(lib="jax_sparse", path="transpose")
        def _(arr, axes=None):
            return bcoo_transpose(arr, permutation=axes)

    except ImportError:
        pass



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Subsystem  class
"""


class Subsystem:
    """A Hilbert space with a name and a dimension."""

    def __init__(self, name: str, dim: int):
        """Initialize with name and dimension.

        Args:
            name: Name of the subsystem.
            dim: Dimension of the subsystem.
        """
        self._name = name
        self._dim = dim

    @property
    def name(self) -> str:
        """Name of subsystem."""
        return self._name

    @property
    def dim(self) -> int:
        """Dimension of subsystem."""
        return self._dim

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Subsystem(name={self.name}, dim={self.dim})"

    def __eq__(self, other: "Subsystem") -> bool:
        if not isinstance(other, Subsystem):
            return False
        return self.name == other.name



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may obtain a copy of this license
# in the LICENSE.txt file in the root directory of this source tree or at
# http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this copyright notice, and modified
# files need to carry a notice indicating that they have been altered from the originals.
# pylint: disable=invalid-name

r"""
Compute perturbation theory terms for an LMDE.
"""

from typing import List, Optional, Callable

from scipy.integrate._ivp.ivp import OdeResult

from multiset import Multiset

from qiskit import QiskitError

from qiskit_dynamics import ArrayLike
from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics.solvers.solver_functions import _is_diffrax_method, _is_jax_method
from qiskit_dynamics.perturbation.multiset_utils import _clean_multisets
from qiskit_dynamics.perturbation.perturbation_utils import (
    _merge_multiset_expansion_order_labels,
    _merge_list_expansion_order_labels,
)

from qiskit_dynamics.perturbation.dyson_magnus import (
    _solve_lmde_dyson,
    _solve_lmde_magnus,
    _solve_lmde_dyson_jax,
    _solve_lmde_magnus_jax,
)


def solve_lmde_perturbation(
    perturbations: List[Callable],
    t_span: ArrayLike,
    expansion_method: str,
    expansion_order: Optional[int] = None,
    expansion_labels: Optional[List[Multiset]] = None,
    perturbation_labels: Optional[List[Multiset]] = None,
    generator: Optional[Callable] = None,
    y0: Optional[ArrayLike] = None,
    dyson_in_frame: Optional[bool] = True,
    integration_method: Optional[str] = "DOP853",
    t_eval: Optional[ArrayLike] = None,
    **kwargs,
) -> OdeResult:
    r"""Compute time-dependent perturbation theory terms for an LMDE.

    This function computes multi-variable Dyson or Magnus expansion terms via the algorithm in
    :footcite:`puzzuoli_algorithms_2023`, or Dyson-like terms via the algorithm in
    :footcite:`haas_engineering_2019`. See the
    :ref:`review on time-dependent perturbation theory <perturbation review>`
    to understand the details and notation used in this documentation.

    Which expansion is used is specified by the ``expansion_method`` argument, which impacts the
    interpretation of several of the function arguments (described below). Regardless of
    ``expansion_method``, the main computation is performed by solving a differential equation,
    utilizing :func:`.solve_ode`, and as such several of the function arguments are direct inputs
    into this function:

        - ``integration_method`` is the ODE method used (passed as ``method`` to
          :func:`.solve_ode`), ``t_span`` is the integration interval, and ``t_eval`` is an optional
          set of points to evaluate the perturbation terms at.
        - ``kwargs`` are passed directly to :func:`.solve_ode`, enabling passing through of
          tolerance or step size arguments.

    Other arguments which are treated the same regardless off ``expansion_method`` are:

        - ``generator`` is the unperturbed generator, and the computation is performed in the
          toggling frame of this generator.

    If ``expansion_method in ['dyson', 'magnus']``, this function computes either multivariable
    Dyson series or Magnus expansion terms. That is, given a (finitely truncated) power series for
    the generator:

    .. math::

        G(t, c_0, \dots, c_{r-1}) = G_\emptyset(t)
            + \sum_{k=1}^\infty \sum_{I \in \mathcal{I}_k(r)} c_I G_I(t),

    this function computes, in the toggling frame of :math:`G_\emptyset(t)` given by ``generator``,
    either a collection of multivariable Dyson terms :math:`\mathcal{D}_I(t)` or multivariable
    Magnus terms :math:`\mathcal{O}_I(t)`, whose definitions are given in the
    :ref:`perturbation theory review <perturbation review>`. In this case, the arguments to the
    function are interpreted as follows:

        - ``perturbations`` and ``perturbation_labels`` specify the truncated generator power
          series. ``perturbations`` provides a list of python callable functions for the non-zero
          :math:`G_I(t)`, and the ``perturbation_labels`` is a list of the corresponding multiset
          labels :math:`I`, in the form of ``Multiset`` instances. If not specified, the labels are
          assumed to be ``[Multiset({0: 1}), ..., Multiset({len(perturbations) - 1: 1})]``.
        - ``expansion_order`` and ``expansion_labels`` specify which terms in the chosen expansion
          are to be computed. ``expansion_order`` specifies that all expansion terms up to a given
          order are to be computed, and ``expansion_labels`` specifies individual terms to be
          computed, specified as ``Multiset`` instances. At least one of ``expansion_order`` and
          ``expansion_labels`` must be specified. If both are specified, then all terms up to
          ``expansion_order`` will be computed, along with any additional specific terms given by
          ``expansion_labels``.

    Note that in the above, this function requires that the Multisets consist of non-negative
    integers. Arguments requiring lists of ``Multiset`` instances also accept lists of any valid
    format acceptable to the ``Multiset`` constructor (modulo the non-negative integer constraint).

    If ``expansion_method == 'dyson_like'``, the setup is different. In this case, for a list of
    matrix-valued functions :math:`G_0(t), \dots, G_{r-1}(t)`, this function computes integrals of
    the form

    .. math::
        \int_{t_0}^{t_F} dt_1 \int_{t_0}^{t_1} dt_2 \dots \int_{t_0}^{t_{k-1}}dt_k
                \tilde{G}_{i_1}(t_1) \dots \tilde{G}_{i_k}(t_k),

    for lists of integers :math:`[i_1, \dots, i_k]`, and similar to the other cases,
    :math:`\tilde{G}_j(t) = V(t)^\dagger G_j(t)V(t)`, i.e. the computation is performed in the
    toggling frame specified by ``generator``.

        - ``perturbations`` gives the list of matrix functions as callables
          :math:`G_0(t), \dots, G_{r-1}(t)`.
        - ``perturbation_labels`` is not used in this mode.
        - ``expansion_order`` specifies that all possible integrals of the above form should be
          computed up to a given order (i.e. integrals up to a given order with all possible
          orderings of the :math:`G_0(t), \dots, G_{r-1}(t)`).
        - ``expansion_labels`` allows for specification of specific terms to be computed. In this
          case, a term is specified by a list of ``int``\s, where the length of the list is the
          order of the integral, and the :math:`G_0(t), \dots, G_{r-1}(t)` appear in the integral in
          the order given by the list.

    Finally, additional optional arguments which can be used in the ``'dyson'`` and ``'dyson_like'``
    cases are:

        - ``dyson_in_frame`` controls which frame the results are returned in. The default is
          ``True``, in which case the results are returned as described above. If ``False``, the
          returned results include a pre-factor of :math:`V(t)`, the solution of the unperturbed
          generator, e.g. :math:`V(t)\mathcal{D}_I(t)`. If ``expansion_method=='magnus'``, this
          argument has no effect on the computation.
        - ``y0`` is the initial state for the LMDE given by the unperturbed generator. The effect of
          this argument on the output is to multiply all outputs by ``y0`` on the right. If ``y0``
          is supplied, the argument ``dyson_in_frame`` must be ``False``, as the operator
          :math:`V(t)` is never explicitly computed, and therefore it cannot be removed from the
          results. If ``y0`` is 1d, it will first be transformed into a 2d column vector to ensure
          consistent usage of matrix multiplication.

    Regardless of the value of ``expansion_method``, results are returned in an ``OdeResult``
    instance in the same manner as :func:`.solve_ode`. The result object stores the solution of the
    LMDE for ``generator`` and ``y0`` as if these were passed directly to :func:`.solve_ode` before,
    as well as the computed perturbation theory terms in the additional attribute
    ``perturbation_data``. If ``expansion_method in ['dyson', 'magnus']``, the ``perturbation_data``
    attribute stores a :class:`.PowerSeriesData` instance, and if
    ``expansion_method == 'dyson_like'``, it stores a :class:`.DysonLikeData` instance. In either
    case, these are data container classes with the following attributes:

        - ``metadata``: Containing expansion information.
        - ``labels``: Index labels for all computed perturbation terms. In the case of the Dyson or
          Magnus expansion, the labels are ``Multiset`` instances, and in the `'dyson_like'` case,
          they are lists of ``int``\s.
        - ``data``: A 4d array storing all computed terms. The first axis indexes the expansion
          terms in the same ordering as ``labels``, and the second axis indexes the perturbation
          terms evaluated at the times in ``results.t`` in the same manner as ``results.y``.

    Additionally, both the :meth:`.PowerSeriesData.get_item` and :meth:`.DysonLikeData.get_item`
    methods can be used to retrieve the results for a given perturbation term according to its
    label. E.g. the results for the term with label ``[0, 1]`` is retrievable via
    ``results.perturbation_data.get_term([0, 1])``.

    Args:
        perturbations: List of matrix-valued callables.
        t_span: Integration bounds.
        expansion_method: Either ``'dyson'``, ``'magnus'``, or ``'dyson_like'``.
        expansion_order: Order of perturbation terms to compute up to. Specifying this argument
            results in computation of all terms up to the given order. Can be used in conjunction
            with ``expansion_labels``.
        expansion_labels: Specific perturbation terms to compute. If both ``expansion_order``
            and ``expansion_labels`` are specified, then all terms up to ``expansion_order`` are
            computed, along with the additional terms specified in ``expansion_labels``.
        perturbation_labels: Optional description of power series terms specified by
            ``perturbations``. To only be used with ``'dyson'`` and ``'magnus'`` methods.
        generator: Optional frame generator. Defaults to ``0``.
        y0: Optional initial state for frame generator LMDE. Defaults to the identity matrix.
        dyson_in_frame: For ``expansion_method`` ``'dyson'`` or ``'dyson_like'``, whether or not
            to remove the frame transformation pre-factor from the Dyson terms.
        integration_method: Integration method to use. Must be supported by :func:`.solve_ode`.
        t_eval: Points at which to evaluate the system.
        **kwargs: Additional arguments to pass to ode integration method used to compute terms.

    Returns:
        OdeResult: Results object containing standard ODE results for LMDE given by ``generator``
        and ``y0``, with additional ``perturbation_data`` attribute containing the requested
        perturbation theory terms.

    Raises:
        QiskitError: If problem with inputs, either ``expansion_method`` is unsupported, or both of
            ``expansion_order`` and ``expansion_labels`` unspecified.

    .. footbibliography::
    """

    # validation checks
    if y0 is not None:
        if "magnus" in expansion_method:
            raise QiskitError("""Argument y0 cannot be used for expansion_method=='magnus'.""")

        if dyson_in_frame:
            raise QiskitError(
                """If expansion_method in ['dyson', 'dyson_like']
                   and y0 passed, dyson_in_frame must be False."""
            )

        # if 1d in a dyson case, turn into a column vector
        if y0.ndim == 1:
            y0 = unp.asarray([y0]).transpose()

    if perturbation_labels is not None and expansion_method == "dyson_like":
        raise QiskitError(
            """perturbation_labels argument not usable with
            expansion_method='dyson_like'."""
        )

    # clean and validate perturbation_labels, and setup expansion terms to compute
    if expansion_method in ["dyson", "magnus"]:
        if perturbation_labels is None:
            perturbation_labels = [Multiset({idx: 1}) for idx in range(len(perturbations))]
        else:
            # validate perturbation_labels
            perturbations_len = len(perturbation_labels)
            perturbation_labels = [Multiset(x) for x in perturbation_labels]
            cleaned_perturbation_labels = _clean_multisets(perturbation_labels)
            if len(cleaned_perturbation_labels) != perturbations_len:
                raise QiskitError("perturbation_labels argument contains duplicates as multisets.")

        expansion_labels = _merge_multiset_expansion_order_labels(
            perturbation_labels=perturbation_labels,
            expansion_order=expansion_order,
            expansion_labels=expansion_labels,
        )
    elif expansion_method in ["dyson_like"]:
        expansion_labels = _merge_list_expansion_order_labels(
            perturbation_num=len(perturbations),
            expansion_order=expansion_order,
            expansion_labels=expansion_labels,
        )

    use_jax = _is_jax_method(integration_method) or _is_diffrax_method(integration_method)

    if expansion_method in ["dyson", "dyson_like"]:
        dyson_like = expansion_method == "dyson_like"
        if not use_jax:
            return _solve_lmde_dyson(
                perturbations=perturbations,
                t_span=t_span,
                dyson_terms=expansion_labels,
                perturbation_labels=perturbation_labels,
                generator=generator,
                y0=y0,
                dyson_in_frame=dyson_in_frame,
                dyson_like=dyson_like,
                integration_method=integration_method,
                t_eval=t_eval,
                **kwargs,
            )
        else:
            return _solve_lmde_dyson_jax(
                perturbations=perturbations,
                t_span=t_span,
                dyson_terms=expansion_labels,
                perturbation_labels=perturbation_labels,
                generator=generator,
                y0=y0,
                dyson_in_frame=dyson_in_frame,
                dyson_like=dyson_like,
                integration_method=integration_method,
                t_eval=t_eval,
                **kwargs,
            )
    elif expansion_method == "magnus":
        if not use_jax:
            return _solve_lmde_magnus(
                perturbations=perturbations,
                t_span=t_span,
                magnus_terms=expansion_labels,
                perturbation_labels=perturbation_labels,
                generator=generator,
                y0=y0,
                integration_method=integration_method,
                t_eval=t_eval,
                **kwargs,
            )
        else:
            return _solve_lmde_magnus_jax(
                perturbations=perturbations,
                t_span=t_span,
                magnus_terms=expansion_labels,
                perturbation_labels=perturbation_labels,
                generator=generator,
                y0=y0,
                integration_method=integration_method,
                t_eval=t_eval,
                **kwargs,
            )

    # raise error if none apply
    raise QiskitError(f"expansion_method {expansion_method} not supported.")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Quantum system model
"""

from typing import Optional, List, Union
from copy import copy

import numpy as np

from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.quantum_info.states.quantum_state import QuantumState

from qiskit_dynamics import ArrayLike
from qiskit_dynamics import Solver, Signal
from qiskit_dynamics.systems import Subsystem, DressedBasis, I, X, N
from qiskit_dynamics.systems.abstract_subsystem_operators import AbstractSubsystemOperator


class QuantumSystemModel:
    """Quantum system model class.

    This class represents an abstract quantum system model containing Hamiltonian and/or Lindblad
    terms, specified in terms of the abstract operator instances provided in this module. Once
    constructed, the :meth:`.get_Solver` method can be used to convert the model into a
    :class:`.Solver` instance with a concrete array representation to solve the system for a given
    initial state. Alternatively, the :meth:`.solve` method can be called to solve the system for an
    initial state without needing to work with the :class:`.Solver` directly. See the :mod:`.models`
    module for a concrete description of the Schrodinger and Lindblad master equations.

    Models can be summed together to build more complex models, e.g. for a system with multiple
    subsystems. See the :ref:`Systems Modelling Tutorial <systems modelling tutorial>` for an
    example of intended usage.
    """

    def __init__(
        self,
        static_hamiltonian: Optional[AbstractSubsystemOperator] = None,
        drive_hamiltonian_coefficients: Optional[List[str]] = None,
        drive_hamiltonians: Optional[List[AbstractSubsystemOperator]] = None,
        static_dissipators: Optional[List[AbstractSubsystemOperator]] = None,
        drive_dissipator_coefficients: Optional[List[str]] = None,
        drive_dissipators: Optional[List[AbstractSubsystemOperator]] = None,
    ):
        """Initialize.

        Args:
            static_hamiltonian: The static Hamiltonian.
            drive_hamiltonian_coefficients: A list of string labels for the drive Hamiltonian terms.
            drive_hamiltonians: The Hamiltonian terms with time-dependent coefficients. This is
                mapped to ``hamiltonian_operators`` in :class:`.Solver`.
            static_dissipators: The static dissipator terms.
            drive_dissipator_coefficients: A list of string labels for the drive dissipator terms.
            drive_dissipators: Dissipator terms with time-dependent rates. This is mapped to
                ``dissipator_operators`` in :class:`.Solver`.
        """

        drive_hamiltonians = drive_hamiltonians or []
        static_dissipators = static_dissipators or []
        drive_dissipators = drive_dissipators or []

        self._operators = {
            "static_hamiltonian": static_hamiltonian,
            "drive_hamiltonians": drive_hamiltonians,
            "static_dissipators": static_dissipators,
            "drive_dissipators": drive_dissipators,
        }

        self._drive_hamiltonian_coefficients = drive_hamiltonian_coefficients or []
        self._drive_dissipator_coefficients = drive_dissipator_coefficients or []

        # is this how we want to make this list?
        subsystems = []
        if static_hamiltonian is not None:
            subsystems = copy(static_hamiltonian.subsystems)
        for op in drive_hamiltonians + static_dissipators + drive_dissipators:
            for x in op.subsystems:
                if x not in subsystems:
                    subsystems.append(x)

        self._subsystems = subsystems

    @property
    def subsystems(self):
        """The model subsystems."""
        return self._subsystems

    @property
    def static_hamiltonian(self):
        """The model static Hamiltonian."""
        return self._operators["static_hamiltonian"]

    @property
    def drive_hamiltonians(self):
        """The model drive Hamiltonians."""
        return self._operators["drive_hamiltonians"]

    @property
    def static_dissipators(self):
        """The model static dissipators."""
        return self._operators["static_dissipators"]

    @property
    def drive_dissipators(self):
        """The model drive dissipators."""
        return self._operators["drive_dissipators"]

    @property
    def drive_hamiltonian_coefficients(self):
        """The drive Hamiltonian coefficients."""
        return self._drive_hamiltonian_coefficients

    @property
    def drive_dissipator_coefficients(self):
        """The drive dissipator coefficients."""
        return self._drive_dissipator_coefficients

    def dressed_basis(self, ordered_subsystems: Optional[List] = None, ordering: str = "default"):
        """Get the DressedBasis object for the system.

        Args:
            ordered_subsystems: Subsystems in the desired order.
            ordering: Ordering convention for the eigenvectors.
        """
        ordered_subsystems = ordered_subsystems or self.subsystems
        return DressedBasis.from_hamiltonian(
            self.static_hamiltonian, ordered_subsystems, ordering=ordering
        )

    def get_Solver(
        self,
        rotating_frame: Optional[Union[np.ndarray, AbstractSubsystemOperator]] = None,
        array_library: Optional[str] = None,
        vectorized: bool = False,
        validate: bool = False,
        ordered_subsystems: Optional[List[Subsystem]] = None,
    ):
        """Build concrete operators and instantiate solver.

        Note that the :meth:`.map_signal_dictionary` method can be used to map signals given in a
        dictionary format ``{drive_coefficient: s}``, where ``drive_coefficient`` is a string in
        ``drive_hamiltonian_coefficients + drive_dissipator_coefficients`` and ``s`` is a signal,
        to the required formatting of the ``signals`` argument in :meth:`.Solver.solve`.

        Args:
            rotating_frame: Rotating frame to define the solver in.
            array_library: array library to use (e.g. "numpy", "jax", "jax_sparse", "scipy_sparse")
            vectorized: If doing lindblad simulation, whether or not to vectorize.
            validate: Whether or not to validate the operators.
            ordered_subsystems: Chosen non-standard ordering for building the solver.
        """
        if ordered_subsystems is None:
            ordered_subsystems = self.subsystems

        if self.static_hamiltonian is None:
            static_hamiltonian = None
        else:
            static_hamiltonian = self.static_hamiltonian.matrix(ordered_subsystems)

        if len(self.drive_hamiltonians) == 0:
            drive_hamiltonians = None
        else:
            drive_hamiltonians = np.array(
                [op.matrix(ordered_subsystems) for op in self.drive_hamiltonians]
            )

        if len(self.static_dissipators) == 0:
            static_dissipators = None
        else:
            static_dissipators = np.array(
                [op.matrix(ordered_subsystems) for op in self.static_dissipators]
            )

        if len(self.static_dissipators) == 0:
            drive_dissipators = None
        else:
            drive_dissipators = np.array(
                [op.matrix(ordered_subsystems) for op in self.drive_dissipators]
            )

        if rotating_frame == "static_hamiltonian":
            rotating_frame = static_hamiltonian
        elif isinstance(rotating_frame, AbstractSubsystemOperator):
            rotating_frame = rotating_frame.matrix(ordered_subsystems)

        return Solver(
            static_hamiltonian=static_hamiltonian,
            hamiltonian_operators=drive_hamiltonians,
            static_dissipators=static_dissipators,
            dissipator_operators=drive_dissipators,
            rotating_frame=rotating_frame,
            validate=validate,
            array_library=array_library,
            vectorized=vectorized,
        )

    def map_signal_dictionary(self, signals: List[Union[ArrayLike, Signal]]):
        """Map labelled signal dictionary to the required format for for the signals argument of a
        :class:`.Solver` generated from the :meth:`.get_Solver` method.

        Args:
            signals: Signals in dictionary format ``{label: s}``, for ``label`` a string in
                ``drive_hamiltonian_coefficients + drive_dissipator_coefficients`` and ``s`` a
                signal.
        Returns:
            A container of signals formatted for the ``signals`` argument of the :class:`.Solver`
            method :meth:`.Solver.solve`.
        """
        # order coefficients
        hamiltonian_signals = [
            signals.get(label, 0.0) for label in self._drive_hamiltonian_coefficients
        ]
        dissipator_signals = [
            signals.get(label, 0.0) for label in self._drive_dissipator_coefficients
        ]

        return (
            hamiltonian_signals
            if len(dissipator_signals) == 0
            else (hamiltonian_signals, dissipator_signals)
        )

    def solve(
        self,
        signals: dict,
        t_span: ArrayLike,
        y0: Union[ArrayLike, QuantumState, BaseOperator],
        rotating_frame: Optional[Union[np.ndarray, AbstractSubsystemOperator]] = None,
        array_library: Optional[str] = None,
        vectorized: Optional[bool] = False,
        ordered_subsystems: Optional[List[Subsystem]] = None,
        **kwargs,
    ):
        """Solve the model.

        This method internally constructs a :class:`.Solver` instance with fully-formed arrays
        according to the abstract model specified in this instance, and then solves. Note that the
        ``signals`` argument for this method expects a dictionary format mapping the coefficient
        labels for the drive terms specified at instantiation to the desired coefficient.

        Args:
            signals: Signals in dictionary format ``{label: s}``, where ``label`` is a string in
                ``drive_hamiltonian_coefficients + drive_dissipator_coefficients``, and ``s`` is the
                corresponding singal.
            t_span: Time interval to integrate over.
            y0: Initial state.
            rotating_frame: Rotating frame to transform the model into. Rotating frames which are
                diagonal can be supplied as a 1d array of the diagonal elements, to explicitly
                indicate that they are diagonal.
            array_library: Array library to use for storing operators of underlying model. See the
                :ref:`model evaluation section of the Models API documentation <model evaluation>`
                for a more detailed description of this argument.
            vectorized: If including dissipator terms, whether or not to construct the
                :class:`.LindbladModel` in vectorized form. See the
                :ref:`model evaluation section of the Models API documentation <model evaluation>`
                for a more detailed description of this argument.
            ordered_subsystems: List of :class:`.Subsystem` instances explicitly specifying the
                ordering of the subsystems desired when building the concrete model.
            kwargs: Keyword arguments to pass through to :class:`.Solver.solve`.
        """

        solver = self.get_Solver(
            rotating_frame=rotating_frame,
            array_library=array_library,
            vectorized=vectorized,
            validate=False,
            ordered_subsystems=ordered_subsystems,
        )

        signals = self.map_signal_dictionary(signals)

        return solver.solve(t_span=t_span, y0=y0, signals=signals, **kwargs)

    def __add__(self, other):
        """Add two models."""

        new_operators = {key: op + other._operators[key] for key, op in self._operators.items()}

        return QuantumSystemModel(
            drive_hamiltonian_coefficients=self.drive_hamiltonian_coefficients
            + other.drive_hamiltonian_coefficients,
            drive_dissipator_coefficients=self.drive_dissipator_coefficients
            + other.drive_dissipator_coefficients,
            **new_operators,
        )

    def __str__(self):
        string = "QuantumSystemModel(\n"
        string += f"    static_hamiltonian={self._operators['static_hamiltonian']},\n"
        string += f"    drive_hamiltonian_coefficients={self.drive_hamiltonian_coefficients},\n"
        string += f"    drive_hamiltonians={self._operators['drive_hamiltonians']},\n"
        string += f"    static_dissipators={self._operators['static_dissipators']},\n"
        string += f"    drive_dissipator_coefficients={self.drive_dissipator_coefficients},\n"
        string += f"    drive_dissipators={self._operators['drive_dissipators']},\n"
        string += ")"

        return string

    def _map_model(self, f):
        """Apply a function to the underlying operators, returning a new QuantumSystemModel.

        If an operator or operator list is ``None``, it will remain ``None`` under the mapping.
        """

        static_hamiltonian = None
        if self.static_hamiltonian is not None:
            static_hamiltonian = f(self.static_hamiltonian)

        drive_hamiltonians = [f(x) for x in self.drive_hamiltonians]
        static_dissipators = [f(x) for x in self.static_dissipators]
        drive_dissipators = [f(x) for x in self.drive_dissipators]

        return QuantumSystemModel(
            static_hamiltonian=static_hamiltonian,
            drive_hamiltonian_coefficients=self.drive_hamiltonian_coefficients,
            drive_hamiltonians=drive_hamiltonians,
            static_dissipators=static_dissipators,
            drive_dissipator_coefficients=self.drive_dissipator_coefficients,
            drive_dissipators=drive_dissipators,
        )


class IdealQubit(QuantumSystemModel):
    r"""Simple dynamical model of a quantum system.

    Intended to represent a 2 level system, though can be constructed on higher dimensional
    subsystems. A model with Hamiltonian of the form :math:`H(t) = 2 \pi \nu Z + s(t) 2 \pi r X`,
    with :math:`\nu` being the frequency, :math:`s(t)` the drive term, and :math:`r` the drive
    strength.
    """

    def __init__(self, subsystem, frequency, drive_strength, drive_label=None):
        """Initialize.

        Args:
            subsystem: The subsystem to define the qubit on.
            frequency: The frequency of the qubit.
            drive_strength: The drive strength of the qubit.
            drive_label: The label for the drive term.
        """
        if drive_label is None:
            drive_label = f"d{subsystem.name}"
        super().__init__(
            static_hamiltonian=2 * np.pi * frequency * N(subsystem),
            drive_hamiltonian_coefficients=[drive_label],
            drive_hamiltonians=[2 * np.pi * drive_strength * X(subsystem) * 0.5],
            static_dissipators=[],
            drive_dissipator_coefficients=[],
            drive_dissipators=[],
        )


class DuffingOscillator(QuantumSystemModel):
    r"""Duffing oscillator.

    A model of a transmon with Hamiltonian:
    :math:`H(t) = 2 \pi \nu N + \pi \alpha N(N - I) + s(t) 2 \pi r X`, where :math:`\nu` is the
    frequency, :math:`\alpha` the anharmonicity, :math:`r` is the drive strength, and :math:`s(t)`
    is the drive signal.
    """

    def __init__(self, subsystem, frequency, anharm, drive_strength, drive_label=None):
        """Initialize.

        Args:
            subsystem: The subsystem to define the Duffing oscillator on.
            frequency: The frequency of the oscillator.
            anharm: The anharmonicity of the oscillator.
            drive_strength: The drive strength.
            drive_label: The label for the drive term.
        """
        if drive_label is None:
            drive_label = f"d{subsystem.name}"

        super().__init__(
            static_hamiltonian=2 * np.pi * frequency * N(subsystem)
            + np.pi * anharm * N(subsystem) @ (N(subsystem) + (-1 * I(subsystem))),
            drive_hamiltonian_coefficients=[drive_label],
            drive_hamiltonians=[2 * np.pi * drive_strength * X(subsystem)],
            static_dissipators=[],
            drive_dissipator_coefficients=[],
            drive_dissipators=[],
        )


class ExchangeInteraction(QuantumSystemModel):
    r"""An exchange interaction between two systems.

    Represents the Hamiltonian :math:`H = g X \otimes X`, where :math:`g` is the strength of the
    coupling, and the two :math:`X` operators act on the two subsystems.
    """

    def __init__(self, subsystems, g):
        """Initialize.

        Args:
            g: The coupling strength.
        """
        super().__init__(
            static_hamiltonian=2 * np.pi * g * (X(subsystems[0]) @ X(subsystems[1])),
            drive_hamiltonian_coefficients=[],
            drive_hamiltonians=[],
            static_dissipators=[],
            drive_dissipator_coefficients=[],
            drive_dissipators=[],
        )



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name, arguments-differ

"""
Pulse-enabled simulator backend.
"""

import datetime
import inspect
import uuid
import warnings

from typing import List, Optional, Union, Dict, Tuple
import copy
import numpy as np
from scipy.integrate._ivp.ivp import OdeResult

from qiskit import pulse
from qiskit.qobj.utils import MeasLevel, MeasReturnType
from qiskit.qobj.common import QobjHeader
from qiskit.transpiler import Target, InstructionProperties
from qiskit.circuit.library import Measure
from qiskit.pulse import Schedule, ScheduleBlock
from qiskit.pulse.transforms.canonicalization import block_to_schedule
from qiskit.providers.options import Options
from qiskit.providers.backend import BackendV2
from qiskit.providers.models.pulsedefaults import PulseDefaults
from qiskit.providers.models.backendconfiguration import PulseBackendConfiguration
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData

from qiskit import QiskitError, QuantumCircuit
from qiskit import schedule as build_schedule
from qiskit.quantum_info import Statevector, DensityMatrix
from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.quantum_info.states.quantum_state import QuantumState

from qiskit_dynamics.arraylias.alias import ArrayLike
from qiskit_dynamics.solvers.solver_classes import Solver

from .dynamics_job import DynamicsJob
from .backend_utils import (
    _get_dressed_state_decomposition,
    _get_lab_frame_static_hamiltonian,
    _get_memory_slot_probabilities,
    _sample_probability_dict,
    _get_counts_from_samples,
    _get_iq_data,
)


class DynamicsBackend(BackendV2):
    r"""Pulse-level simulator backend.

    This class provides a :class:`~qiskit.providers.backend.BackendV2` interface wrapper around a
    :class:`.Solver` instance setup to simulate pulse schedules. The backend can be configured to
    take advantage of standard transpilation infrastructure to describe pulse-level simulations in
    terms of :class:`~qiskit.circuit.QuantumCircuit`\s. Results are returned as
    :class:`~qiskit.result.Result` instances.

    A minimal :class:`.DynamicsBackend` requires specifying only a :class:`.Solver` instance and a
    list of subsystem dimensions, indicating the subsystem decomposition of the model in
    :class:`.Solver`. For example, the following code builds a :class:`.DynamicsBackend` around a
    :class:`.Solver` and indicates that the system specified by the :class:`.Solver` decomposes as
    two ``3`` dimensional subsystems.

    .. code-block:: python

        backend = DynamicsBackend(
            solver=solver, subsystem_dims=[3, 3]
        )

    Without further configuration, the above ``backend`` can be used to simulate either
    :class:`~qiskit.pulse.Schedule` or :class:`~qiskit.pulse.ScheduleBlock` instances.

    Pulse-level simulations defined in terms of :class:`~qiskit.circuit.QuantumCircuit` instances
    can also be performed if each gate in the circuit has a corresponding pulse-level definition,
    either as an attached calibration, or as an instruction contained in ``backend.target``.


    **Supported options**

    The behaviour of the backend can be configured via the following options. These can either be
    passed as optional keyword arguments at construction, set with the
    :meth:`.DynamicsBackend.set_options` method after construction, or passed as runtime arguments
    to :meth:`.DynamicsBackend.run`.

    * ``shots``: Number of shots per experiment. Defaults to ``1024``.
    * ``solver``: The Qiskit Dynamics :class:`.Solver` instance used for simulation.
    * ``solver_options``: Dictionary containing optional kwargs for passing to :meth:`Solver.solve`,
      indicating solver methods and options. Defaults to the empty dictionary ``{}``.
    * ``subsystem_dims``: Dimensions of subsystems making up the system in ``solver``. Defaults to
      ``[solver.model.dim]``.
    * ``meas_map``: Measurement map. Defaults to ``[[idx] for idx in range(len(subsystem_dims))]``.
    * ``control_channel_map``: A dictionary mapping control channel labels to indices, to be used
      for control channel index lookup in the :meth:`DynamicsBackend.control_channel` method.
    * ``initial_state``: Initial state for simulation, either the string ``"ground_state"``,
      indicating that the ground state for the system Hamiltonian should be used, or an arbitrary
      ``Statevector`` or ``DensityMatrix``. Defaults to ``"ground_state"``.
    * ``normalize_states``: Boolean indicating whether to normalize states before computing outcome
      probabilities, and normalize probablities before sampling. Defaults to ``True``. Setting to
      ``False`` can result in errors if the solution tolerance results in probabilities with
      significant numerical deviation from a proper probability distribution.
    * ``meas_level``: Form of measurement output. Supported values are ``1`` and ``2``. ``1``
      returns IQ points and ``2`` returns counts. Defaults to ``meas_level == 2``.
    * ``meas_return``: Level of measurement data to return. For ``meas_level = 1`` ``"single"``
      returns output from every shot. ``"avg"`` returns average over shots of measurement output.
      Defaults to ``"avg"``.
    * ``iq_centers``: Centers for IQ distribution when using ``meas_level==1`` results. Must have
      type ``List[List[List[float, float]]]`` formatted as ``iq_centers[subsystem][level] = [I,
      Q]``. If ``None``, the ``iq_centers`` are dynamically generated to be equally spaced points on
      a unit circle with ground-state at ``(1, 0)``. The default is ``None``.
    * ``iq_width``: Standard deviation of IQ distribution around the centers for ``meas_level==1``.
      Must be a positive float. Defaults to ``0.2``.
    * ``max_outcome_level``: For ``meas_level == 2``, the maximum outcome for each subsystem. Values
      will be rounded down to be no larger than ``max_outcome_level``. Must be a positive integer or
      ``None``. If ``None``, defaults to ``1``.
    * ``memory``: Boolean indicating whether to return a list of explicit measurement outcomes for
      every experimental shot. Defaults to ``True``.
    * ``seed_simulator``: Seed to use in random sampling. Defaults to ``None``.
    * ``experiment_result_function``: Function for computing the ``ExperimentResult`` for each
      simulated experiment. This option defaults to :func:`default_experiment_result_function`, and
      any other function set to this option must have the same signature. Note that the default
      utilizes various other options that control results computation, and hence changing it will
      impact the meaning of other options.
    * ``configuration``: A :class:`PulseBackendConfiguration` instance or ``None``. This option
      defaults to ``None``, and is not required for the functioning of this class, but is provided
      for compatibility. A set configuration will be returned by
      :meth:`DynamicsBackend.configuration()`.
    * ``defaults``: A :class:`PulseDefaults` instance or ``None``. This option defaults to ``None``,
      and is not required for the functioning of this class, but is provided for compatibility. A
      set defaults will be returned by :meth:`DynamicsBackend.defaults()`.
    """

    def __init__(
        self,
        solver: Solver,
        target: Optional[Target] = None,
        **options,
    ):
        """Instantiate with a :class:`.Solver` instance and additional options.

        Args:
            solver: Solver instance configured for pulse simulation.
            target: Target object.
            options: Additional configuration options for the simulator.

        Raises:
            QiskitError: If any instantiation arguments fail validation checks.
        """

        super().__init__(
            name="DynamicsBackend",
            description="Pulse enabled simulator backend.",
            backend_version="0.1",
        )

        # Dressed states of solver, will be calculated when solver option is set
        self._dressed_evals = None
        self._dressed_states = None
        self._dressed_states_adjoint = None

        # add subsystem_dims to options so set_options validation works
        if "subsystem_dims" not in options:
            options["subsystem_dims"] = [solver.model.dim]

        # Set simulator options
        self.set_options(solver=solver, **options)

        if self.options.meas_map is None:
            meas_map = [[idx] for idx in range(len(self.options.subsystem_dims))]
            self.set_options(meas_map=meas_map)

        # self._target = target or Target() doesn't work as bool(target) can be False
        if target is None:
            target = Target()
        else:
            target = copy.copy(target)

        # add default simulator measure instructions
        measure_properties = {}
        instruction_schedule_map = target.instruction_schedule_map()
        for qubit in range(len(self.options.subsystem_dims)):
            if not instruction_schedule_map.has(instruction="measure", qubits=qubit):
                with pulse.build() as meas_sched:
                    pulse.acquire(
                        duration=1, qubit_or_channel=qubit, register=pulse.MemorySlot(qubit)
                    )

                measure_properties[(qubit,)] = InstructionProperties(calibration=meas_sched)

        if bool(measure_properties):
            target.add_instruction(Measure(), measure_properties)

        target.dt = solver._dt

        self._target = target

    def _default_options(self):
        return Options(
            shots=1024,
            solver=None,
            solver_options={},
            subsystem_dims=None,
            meas_map=None,
            control_channel_map=None,
            normalize_states=True,
            initial_state="ground_state",
            meas_level=MeasLevel.CLASSIFIED,
            meas_return=MeasReturnType.AVERAGE,
            iq_centers=None,
            iq_width=0.2,
            max_outcome_level=1,
            memory=True,
            seed_simulator=None,
            experiment_result_function=default_experiment_result_function,
            configuration=None,
            defaults=None,
        )

    def set_options(self, **fields):
        """Set options for DynamicsBackend."""

        validate_subsystem_dims = False
        validate_iq_centers = False

        for key, value in fields.items():
            if not hasattr(self._options, key):
                raise AttributeError(f"Invalid option {key}")

            # validation checks
            if key == "initial_state":
                if value != "ground_state" and not isinstance(value, (Statevector, DensityMatrix)):
                    raise QiskitError(
                        'initial_state must be either "ground_state", or a Statevector or '
                        "DensityMatrix instance."
                    )
            elif key == "meas_level" and value not in [1, 2]:
                raise QiskitError("Only meas_level 1 and 2 are supported by DynamicsBackend.")
            elif key == "meas_return" and value not in ["single", "avg"]:
                raise QiskitError("meas_return must be either 'single' or 'avg'.")
            elif key == "max_outcome_level":
                if (value is not None) and (not isinstance(value, int) or (value <= 0)):
                    raise QiskitError("max_outcome_level must be a positive integer or None.")
            elif key == "experiment_result_function" and not callable(value):
                raise QiskitError("experiment_result_function must be callable.")
            elif key == "configuration" and not isinstance(value, PulseBackendConfiguration):
                raise QiskitError(
                    "configuration option must be an instance of PulseBackendConfiguration."
                )
            elif key == "defaults" and not isinstance(value, PulseDefaults):
                raise QiskitError("defaults option must be an instance of PulseDefaults.")
            elif key == "iq_width" and (not isinstance(value, float) or (value <= 0)):
                raise QiskitError("iq_width must be a positive float.")
            elif key == "iq_centers":
                if (value is not None) and not all(
                    (isinstance(level, List) and len(level) == 2)
                    for sub_system in value
                    for level in sub_system
                ):
                    raise QiskitError(
                        "The iq_centers option must be either None or of type "
                        "List[List[List[int, int]]], where the innermost list is the (I, Q) pair."
                    )
                validate_iq_centers = True
            elif key == "subsystem_dims":
                validate_subsystem_dims = True
                validate_iq_centers = True
            elif key == "solver":
                validate_subsystem_dims = True
            elif key == "control_channel_map":
                if value is not None:
                    if not isinstance(value, dict):
                        raise QiskitError(
                            "The control_channel_map option must either be None or a dictionary."
                        )
                    if not all(isinstance(x, int) for x in value.values()):
                        raise QiskitError("The control_channel_map values must be of type int.")

            # special setting routines
            if key == "solver":
                self._set_solver(value)
            else:
                self._options.update_options(**{key: value})

        # perform additional consistency validations if certain options were modified
        if (
            validate_subsystem_dims
            and np.prod(self._options.subsystem_dims) != self._options.solver.model.dim
        ):
            raise QiskitError(
                "DynamicsBackend options subsystem_dims and solver.model.dim are inconsistent."
            )

        if validate_iq_centers and (self._options.iq_centers is not None):
            if [
                len(sub_system) for sub_system in self._options.iq_centers
            ] != self._options.subsystem_dims:
                raise QiskitError(
                    """iq_centers option is not consistent with subsystem_dims. Must be None
                or of type List[List[List[int, int]]], where the outermost list is of length equal
                to the number of subsystems, and each inner list of length equal to the
                corresponding subsystem dimension."""
                )

    def _set_solver(self, solver):
        """Configure simulator based on provided solver."""
        if solver._dt is None:
            raise QiskitError(
                "Solver passed to DynamicsBackend is not configured for Pulse simulation."
            )

        self._options.update_options(solver=solver)
        # Get dressed states
        static_hamiltonian = _get_lab_frame_static_hamiltonian(solver.model)
        dressed_evals, dressed_states = _get_dressed_state_decomposition(static_hamiltonian)
        self._dressed_evals = dressed_evals
        self._dressed_states = dressed_states
        self._dressed_states_adjoint = self._dressed_states.conj().transpose()

    def solve(
        self,
        solve_input: List[Union[QuantumCircuit, Schedule, ScheduleBlock]],
        t_span: Optional[ArrayLike] = None,
        y0: Optional[Union[ArrayLike, QuantumState, BaseOperator]] = None,
        convert_results: Optional[bool] = True,
        validate: Optional[bool] = True,
    ) -> Union[OdeResult, List[OdeResult]]:
        """Simulate a list of :class:`~qiskit.circuit.QuantumCircuit`,
        :class:`~qiskit.pulse.Schedule`, or :class:`~qiskit.pulse.ScheduleBlock` instances and
        return the ``OdeResult``.

        This method is analogous to :meth:`.Solver.solve`, however it additionally utilizes
        transpilation and the backend configuration to convert
        :class:`~qiskit.circuit.QuantumCircuit` instances into pulse-level schedules for simulation.
        The options for the solver will be drawn from ``self.options.solver_options``, and if
        ``y0`` is not specified, it will be set from ``self.options.initial_state``.

        Args:
            t_span: Time interval to integrate over. Defaults to ``None``, in which case the
                    interval is set to ``[[0, input.duration] for input in solve_input]]``.
            y0: Initial state.
            solve_input: Time evolution of the system in terms of quantum circuits or qiskit
                         pulse schedules.
            convert_results: If ``True``, convert returned solver state results to the same class
                             as y0. If ``False``, states will be returned in the native array type
                             used by the specified solver method.
            validate: Whether or not to run validation checks on the input.

        Returns:
            OdeResult: object with formatted output types.
        """
        if validate:
            _validate_run_input(solve_input)
        schedules, _ = _to_schedule_list(solve_input, backend=self)

        # use default y0 if not given as parameter
        if y0 is None:
            y0 = self.options.initial_state
        if isinstance(y0, str) and y0 == "ground_state":
            y0 = Statevector(self._dressed_states[:, 0])
        if t_span is None:
            t_span = [[0, sched.duration * self.dt] for sched in schedules]
        solver_results = self.options.solver.solve(
            t_span=t_span,
            y0=y0,
            signals=schedules,
            convert_results=convert_results,
            **self.options.solver_options,
        )
        return solver_results

    # pylint: disable=arguments-differ
    def run(
        self,
        run_input: List[Union[QuantumCircuit, Schedule, ScheduleBlock]],
        validate: Optional[bool] = True,
        **options,
    ) -> DynamicsJob:
        """Run a list of simulations.

        Args:
            run_input: A list of simulations, specified by ``QuantumCircuit``, ``Schedule``, or
                ``ScheduleBlock`` instances.
            validate: Whether or not to run validation checks on the input.
            **options: Additional run options to temporarily override current backend options.

        Returns:
            DynamicsJob object containing results and status.

        Raises:
            QiskitError: If invalid options are set.
        """

        if validate:
            _validate_run_input(run_input)

        # Configure run options for simulation
        if options:
            backend = copy.deepcopy(self)
            backend.set_options(**options)
        else:
            backend = self

        schedules, num_memory_slots_list = _to_schedule_list(run_input, backend=backend)

        # get the acquires sample times and subsystem measurement information
        (
            t_span,
            measurement_subsystems_list,
            memory_slot_indices_list,
        ) = _get_acquire_instruction_timings(
            schedules, backend.options.subsystem_dims, backend.options.solver._dt
        )

        # Build and submit job
        job_id = str(uuid.uuid4())
        dynamics_job = DynamicsJob(
            backend=backend,
            job_id=job_id,
            fn=backend._run,
            fn_kwargs={
                "t_span": t_span,
                "schedules": schedules,
                "measurement_subsystems_list": measurement_subsystems_list,
                "memory_slot_indices_list": memory_slot_indices_list,
                "num_memory_slots_list": num_memory_slots_list,
            },
        )
        dynamics_job.submit()

        return dynamics_job

    def _run(
        self,
        job_id,
        t_span,
        schedules,
        measurement_subsystems_list,
        memory_slot_indices_list,
        num_memory_slots_list,
    ) -> Result:
        """Simulate a list of schedules."""

        # simulate all schedules
        y0 = self.options.initial_state
        if y0 == "ground_state":
            y0 = Statevector(self._dressed_states[:, 0])

        solver_results = self.options.solver.solve(
            t_span=t_span, y0=y0, signals=schedules, **self.options.solver_options
        )

        # compute results for each experiment
        experiment_names = [schedule.name for schedule in schedules]
        experiment_metadatas = [schedule.metadata for schedule in schedules]
        rng = np.random.default_rng(self.options.seed_simulator)
        experiment_results = []
        for (
            experiment_name,
            solver_result,
            measurement_subsystems,
            memory_slot_indices,
            num_memory_slots,
            experiment_metadata,
        ) in zip(
            experiment_names,
            solver_results,
            measurement_subsystems_list,
            memory_slot_indices_list,
            num_memory_slots_list,
            experiment_metadatas,
        ):
            experiment_results.append(
                self.options.experiment_result_function(
                    experiment_name,
                    solver_result,
                    measurement_subsystems,
                    memory_slot_indices,
                    num_memory_slots,
                    self,
                    seed=rng.integers(low=0, high=9223372036854775807),
                    metadata=experiment_metadata,
                )
            )

        # Construct full result object
        return Result(
            backend_name=self.name,
            backend_version=self.backend_version,
            qobj_id="",
            job_id=job_id,
            success=True,
            results=experiment_results,
            date=datetime.datetime.now().isoformat(),
        )

    @property
    def max_circuits(self):
        return None

    @property
    def target(self) -> Target:
        return self._target

    @property
    def meas_map(self) -> List[List[int]]:
        return self.options.meas_map

    def _get_qubit_channel(
        self, qubit: int, ChannelClass: pulse.channels.Channel, method_name: str
    ):
        """Construct a channel instance for a given qubit."""
        if qubit < len(self.options.subsystem_dims):
            return ChannelClass(qubit)

        raise QiskitError(f"{method_name} requested for qubit {qubit}, which is out of bounds.")

    def drive_channel(self, qubit: int) -> pulse.DriveChannel:
        """Return the drive channel for a given qubit."""
        return self._get_qubit_channel(qubit, pulse.DriveChannel, "drive_channel")

    def measure_channel(self, qubit: int) -> pulse.MeasureChannel:
        """Return the measure channel for a given qubit."""
        return self._get_qubit_channel(qubit, pulse.MeasureChannel, "measure_channel")

    def acquire_channel(self, qubit: int) -> pulse.AcquireChannel:
        """Return the measure channel for a given qubit."""
        return self._get_qubit_channel(qubit, pulse.AcquireChannel, "acquire_channel")

    def control_channel(
        self, qubits: Union[Tuple[int, int], List[Tuple[int, int]]]
    ) -> List[pulse.ControlChannel]:
        """Return the control channel with a given label specified by qubits.

        This method requires the ``control_channel_map`` option is set, and otherwise will raise
        a ``NotImplementedError``.

        Args:
            qubits: The label for the control channel, or a list of labels.
        Returns:
            A list containing the control channels specified by qubits.
        Raises:
            NotImplementedError: If the control_channel_map option is not set for this backend.
            QiskitError: If a requested channel is not in the control_channel_map.
        """
        if self.options.control_channel_map is None:
            raise NotImplementedError

        if not isinstance(qubits, list):
            qubits = [qubits]

        control_channels = []
        for x in qubits:
            if x not in self.options.control_channel_map:
                raise QiskitError(f"Key {x} not in control_channel_map.")
            control_channels.append(pulse.ControlChannel(self.options.control_channel_map[x]))

        return control_channels

    def configuration(self) -> PulseBackendConfiguration:
        """Get the backend configuration."""
        return self.options.configuration

    def defaults(self) -> PulseDefaults:
        """Get the backend defaults."""
        return self.options.defaults


def default_experiment_result_function(
    experiment_name: str,
    solver_result: OdeResult,
    measurement_subsystems: List[int],
    memory_slot_indices: List[int],
    num_memory_slots: Union[None, int],
    backend: DynamicsBackend,
    seed: Optional[int] = None,
    metadata: Optional[Dict] = None,
) -> ExperimentResult:
    """Default routine for generating ExperimentResult object.

    To generate the results for a given experiment, this method takes the following steps:

    * The final state is transformed out of the rotating frame and into the lab frame using
      ``backend.options.solver``.
    * If ``backend.options.normalize_states==True``, the final state is normalized.
    * Measurement results are computed, in the dressed basis, based on both the measurement-related
      options in ``backend.options`` and the measurement specification extracted from the specific
      experiment.

    Args:
        experiment_name: Name of experiment.
        solver_result: Result object from :class:`Solver.solve`.
        measurement_subsystems: Labels of subsystems in the model being measured.
        memory_slot_indices: Indices of memory slots to store the results in for each subsystem.
        num_memory_slots: Total number of memory slots in the returned output. If ``None``,
            ``max(memory_slot_indices)`` will be used.
        backend: The backend instance that ran the simulation. Various options and properties
            are utilized.
        seed: Seed for any random number generation involved (e.g. when computing outcome samples).
        metadata: Metadata to add to the header of the
            :class:`~qiskit.result.models.ExperimentResult` object.

    Returns:
        :class:`~qiskit.result.models.ExperimentResult` object containing results.

    Raises:
        QiskitError: If a specified option is unsupported.
    """

    yf = solver_result.y[-1]
    tf = solver_result.t[-1]

    # Take state out of frame, put in dressed basis, and normalize
    if isinstance(yf, Statevector):
        yf = np.array(backend.options.solver.model.rotating_frame.state_out_of_frame(t=tf, y=yf))
        yf = backend._dressed_states_adjoint @ yf
        yf = Statevector(yf, dims=backend.options.subsystem_dims)

        if backend.options.normalize_states:
            yf = yf / np.linalg.norm(yf.data)
    elif isinstance(yf, DensityMatrix):
        yf = np.array(
            backend.options.solver.model.rotating_frame.operator_out_of_frame(t=tf, operator=yf)
        )
        yf = backend._dressed_states_adjoint @ yf @ backend._dressed_states
        yf = DensityMatrix(yf, dims=backend.options.subsystem_dims)

        if backend.options.normalize_states:
            yf = yf / np.diag(yf.data).sum()

    if backend.options.meas_level == MeasLevel.CLASSIFIED:
        memory_slot_probabilities = _get_memory_slot_probabilities(
            probability_dict=yf.probabilities_dict(qargs=measurement_subsystems),
            memory_slot_indices=memory_slot_indices,
            num_memory_slots=num_memory_slots,
            max_outcome_value=backend.options.max_outcome_level,
        )

        # sample
        memory_samples = _sample_probability_dict(
            memory_slot_probabilities,
            shots=backend.options.shots,
            normalize_probabilities=backend.options.normalize_states,
            seed=seed,
        )
        counts = _get_counts_from_samples(memory_samples)

        # construct results object
        exp_data = ExperimentResultData(
            counts=counts, memory=memory_samples.tolist() if backend.options.memory else None
        )
        return ExperimentResult(
            shots=backend.options.shots,
            success=True,
            data=exp_data,
            meas_level=MeasLevel.CLASSIFIED,
            seed=seed,
            header=QobjHeader(name=experiment_name, metadata=metadata),
        )
    elif backend.options.meas_level == MeasLevel.KERNELED:
        iq_centers = backend.options.iq_centers
        if iq_centers is None:
            # Default iq_centers
            iq_centers = []
            for sub_dim in backend.options.subsystem_dims:
                theta = 2 * np.pi / sub_dim
                iq_centers.append(
                    [(np.cos(idx * theta), np.sin(idx * theta)) for idx in range(sub_dim)]
                )

        # generate IQ
        measurement_data = _get_iq_data(
            yf,
            measurement_subsystems=measurement_subsystems,
            iq_centers=iq_centers,
            iq_width=backend.options.iq_width,
            shots=backend.options.shots,
            memory_slot_indices=memory_slot_indices,
            num_memory_slots=num_memory_slots,
            seed=seed,
        )

        if backend.options.meas_return == MeasReturnType.AVERAGE:
            measurement_data = np.average(measurement_data, axis=0)

        # construct results object
        exp_data = ExperimentResultData(memory=measurement_data.tolist())
        return ExperimentResult(
            shots=backend.options.shots,
            success=True,
            data=exp_data,
            meas_level=MeasLevel.KERNELED,
            seed=seed,
            header=QobjHeader(name=experiment_name, metadata=metadata),
        )

    else:
        raise QiskitError(f"meas_level=={backend.options.meas_level} not implemented.")


def _validate_run_input(run_input, accept_list=True, caller_func_name: str = None):
    """Raise errors if the run_input is not one of QuantumCircuit, Schedule, ScheduleBlock, or
    a list of these.
    """
    if caller_func_name is None:
        caller_func_name = inspect.stack()[1].function
    if isinstance(run_input, list) and accept_list:
        # if list apply recursively, but no longer accept lists
        for x in run_input:
            _validate_run_input(x, accept_list=False, caller_func_name=caller_func_name)
    elif not isinstance(run_input, (QuantumCircuit, Schedule, ScheduleBlock)):
        raise QiskitError(
            f"Input type {type(run_input)} not supported by DynamicsBackend.{caller_func_name}."
        )


def _get_acquire_instruction_timings(
    schedules: List[Schedule], subsystem_dims: List[int], dt: float
) -> Tuple[List[List[float]], List[List[int]], List[List[int]]]:
    """Get the required data from the acquire commands in each schedule.

    Additionally validates that each schedule has Acquire instructions occurring at one time, at
    least one memory slot is being listed, and all measured subsystem indices are less than
    ``len(subsystem_dims)``. Additionally, a warning is raised if a 'trivial' subsystem is measured,
    i.e. one with dimension 1.

    Args:
        schedules: A list of schedules.
        subsystem_dims: List of subsystem dimensions.
        dt: The sample size.
    Returns:
        A tuple of lists containing, for each schedule: the list of integration intervals required
        for each schedule (in absolute time, from 0.0 to the beginning of the acquire instructions),
        a list of the subsystems being measured, and a list of the memory slots indices in which to
        store the results of each subsystem measurement.
    Raises:
        QiskitError: If a schedule contains no measurement, if a schedule contains measurements at
            different times, or if a measurement has an invalid subsystem label.
    """
    t_span_list = []
    measurement_subsystems_list = []
    memory_slot_indices_list = []
    for schedule in schedules:
        schedule_acquires = []
        schedule_acquire_times = []
        for start_time, inst in schedule.instructions:
            # only track acquires saving in a memory slot
            if isinstance(inst, pulse.Acquire) and inst.mem_slot is not None:
                schedule_acquires.append(inst)
                schedule_acquire_times.append(start_time)

        # validate
        if len(schedule_acquire_times) == 0:
            raise QiskitError(
                "At least one measurement saving a a result in a MemorySlot "
                "must be present in each schedule."
            )

        for acquire_time in schedule_acquire_times[1:]:
            if acquire_time != schedule_acquire_times[0]:
                raise QiskitError("DynamicsBackend.run only supports measurements at one time.")

        # use dt to convert acquire start time from sample index to the integration interval
        t_span_list.append([0.0, dt * schedule_acquire_times[0]])
        measurement_subsystems = []
        memory_slot_indices = []
        for inst in schedule_acquires:
            if not inst.channel.index < len(subsystem_dims):
                raise QiskitError(
                    f"Attempted to measure out of bounds subsystem {inst.channel.index}."
                )

            if subsystem_dims[inst.channel.index] == 1:
                warnings.warn(f"Measuring trivial subsystem {inst.channel.index} with dimension 1.")

            measurement_subsystems.append(inst.channel.index)

            memory_slot_indices.append(inst.mem_slot.index)

        measurement_subsystems_list.append(measurement_subsystems)
        memory_slot_indices_list.append(memory_slot_indices)

    return t_span_list, measurement_subsystems_list, memory_slot_indices_list


def _to_schedule_list(
    run_input: List[Union[QuantumCircuit, Schedule, ScheduleBlock]], backend: BackendV2
):
    """Convert all inputs to schedules, and store the number of classical registers present
    in any circuits.
    """
    if not isinstance(run_input, list):
        run_input = [run_input]

    schedules = []
    num_memslots = []
    for sched in run_input:
        num_memslots.append(None)
        if isinstance(sched, ScheduleBlock):
            schedules.append(block_to_schedule(sched))
        elif isinstance(sched, Schedule):
            schedules.append(sched)
        elif isinstance(sched, QuantumCircuit):
            num_memslots[-1] = sum(creg.size for creg in sched.cregs)
            schedules.append(build_schedule(sched, backend, dt=backend.options.solver._dt))
        else:
            raise QiskitError(f"Type {type(sched)} cannot be converted to Schedule.")
    return schedules, num_memslots


def _get_backend_channel_freqs(
    backend_target: Optional[Target],
    backend_config: PulseBackendConfiguration,
    backend_defaults: Optional[PulseDefaults],
    channels: List[str],
) -> Dict[str, float]:
    """Extract frequencies of channels from a backend configuration and defaults.

    Args:
        backend_target: A backend target object or ``None``.
        backend_config: A backend configuration object.
        backend_defaults: A backend defaults object or ``None``.
        channels: Channel labels given as strings, assumed to be unique.

    Returns:
        Dict: Mapping of channel labels to frequencies.

    Raises:
        QiskitError: If the frequency for one of the channels cannot be found.
    """

    # partition types of channels
    drive_channels = []
    meas_channels = []
    u_channels = []

    for channel in channels:
        if channel[0] == "d":
            drive_channels.append(channel)
        elif channel[0] == "m":
            meas_channels.append(channel)
        elif channel[0] == "u":
            u_channels.append(channel)
        else:
            raise QiskitError("Unrecognized channel type requested.")

    # extract and validate channel frequency parameters
    if drive_channels:
        # get drive channel frequencies
        drive_frequencies = []
        if (backend_target is not None) and (backend_target.qubit_properties is not None):
            drive_frequencies = [q.frequency for q in backend_target.qubit_properties]
        elif backend_defaults is not None:
            drive_frequencies = backend_defaults.qubit_freq_est
        else:
            raise QiskitError(
                "DriveChannels in model but frequencies not available in target or defaults."
            )

    if meas_channels:
        if backend_defaults is not None:
            meas_frequencies = backend_defaults.meas_freq_est
        else:
            raise QiskitError("MeasureChannels in model but defaults does not have meas_freq_est.")

    # backend_config.u_channel_lo is guaranteed to be a list
    u_channel_lo = backend_config.u_channel_lo

    # populate frequencies
    channel_freqs = {}

    for channel in drive_channels:
        idx = int(channel[1:])
        # pylint: disable=possibly-used-before-assignment
        if idx >= len(drive_frequencies):
            raise QiskitError(f"DriveChannel index {idx} is out of bounds.")
        channel_freqs[channel] = drive_frequencies[idx]

    for channel in meas_channels:
        idx = int(channel[1:])
        # pylint: disable=possibly-used-before-assignment
        if idx >= len(meas_frequencies):
            raise QiskitError(f"MeasureChannel index {idx} is out of bounds.")
        channel_freqs[channel] = meas_frequencies[idx]

    for channel in u_channels:
        idx = int(channel[1:])
        if idx >= len(u_channel_lo):
            raise QiskitError(f"ControlChannel index {idx} is out of bounds.")
        freq = 0.0
        for channel_lo in u_channel_lo[idx]:
            freq += drive_frequencies[channel_lo.q] * channel_lo.scale

        channel_freqs[channel] = freq

    # validate that all channels have frequencies
    for channel in channels:
        if channel not in channel_freqs:
            raise QiskitError(f"No carrier frequency found for channel {channel}.")

    return channel_freqs



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Utility functions for solvers.
"""

from typing import Optional, List, Tuple, Callable
import numpy as np
from scipy.integrate._ivp.ivp import OdeResult

from qiskit import QiskitError

from qiskit_dynamics.arraylias import ArrayLike
from qiskit_dynamics.models import LindbladModel

try:
    from jax.lax import cond
    import jax.numpy as jnp
except ImportError:
    pass


def is_lindblad_model_vectorized(obj: any) -> bool:
    """Return True if obj is a vectorized LindbladModel."""
    return isinstance(obj, LindbladModel) and obj.vectorized


def is_lindblad_model_not_vectorized(obj: any) -> bool:
    """Return True if obj is a non-vectorized LindbladModel."""
    return isinstance(obj, LindbladModel) and not obj.vectorized


def merge_t_args(t_span: ArrayLike, t_eval: Optional[ArrayLike] = None) -> np.ndarray:
    """Merge ``t_span`` and ``t_eval`` into a single array.

    Validition is similar to scipy ``solve_ivp``: ``t_eval`` must be contained in ``t_span``, and be
    increasing if ``t_span[1] > t_span[0]`` or decreasing if ``t_span[1] < t_span[0]``.

    Note: this is done explicitly with ``numpy``, and hence this is not differentiable or compilable
    using jax.

    If ``t_eval is None`` returns ``t_span`` with no modification.

    Args:
        t_span: Interval to solve over.
        t_eval: Time points to include in returned results.

    Returns:
        np.ndarray: Combined list of times.

    Raises:
        ValueError: If one of several validation checks fail.
    """

    if t_eval is None:
        return t_span

    t_span = np.array(t_span)

    t_min = np.min(t_span)
    t_max = np.max(t_span)
    t_direction = np.sign(t_span[1] - t_span[0])

    t_eval = np.array(t_eval)

    if t_eval.ndim > 1:
        raise ValueError("t_eval must be 1 dimensional.")

    if np.min(t_eval) < t_min or np.max(t_eval) > t_max:
        raise ValueError("t_eval entries must lie in t_span.")

    diff = np.diff(t_eval)

    if np.any(t_direction * diff < 0.0):
        raise ValueError("t_eval must be ordered according to the direction of integration.")

    # add endpoints
    t_eval = np.append(np.append(t_span[0], t_eval), t_span[1])

    return t_eval


def trim_t_results(
    results: OdeResult,
    t_eval: Optional[ArrayLike] = None,
) -> OdeResult:
    """Trim ``OdeResult`` object if ``t_eval is not None``.

    Args:
        results: Result object, assumed to contain solution at time points
                 from the output of ``merge_t_args(t_span, t_eval)``.
        t_eval: Time points to include in returned results.

    Returns:
        OdeResult: Results with only times/solutions in ``t_eval``. If ``t_eval``
                   is ``None``, does nothing, returning solver default output.
    """

    if t_eval is None:
        return results

    # remove endpoints
    results.t = results.t[1:-1]
    results.y = results.y[1:-1]

    return results


def merge_t_args_jax(t_span: ArrayLike, t_eval: Optional[ArrayLike] = None) -> np.ndarray:
    """JAX-compilable version of merge_t_args.

    Rather than raise errors, sets return values to ``jnp.nan`` to signal errors.
    The merging strategy differs from :func:`merge_t_args`: after appending
    the endpoints of ``t_span`` to ``t_eval``, it is checked whether the first two entries
    of the resulting array are equal. If they are, the second entry is set to the average
    of the first and the third. A similar procedure is done for the last two entries. This is
    essentially a hack to to avoid adjacent entries of the combined ``t_span``, ``t_eval``
    array having equal values, which causes buggy behaviour in :func:`jax_odeint`.

    Args:
        t_span: Interval to solve over. Assumed to be a list, tuple, or Array with 2 entries.
        t_eval: Time points to include in returned results.

    Returns:
        np.ndarray: Combined list of times.

    Raises:
        ValueError: If either argument is not one dimensional.
    """

    if t_eval is None:
        return jnp.array(t_span)

    t_span = jnp.array(t_span)
    t_eval = jnp.array(t_eval)

    # raise error if not one dimensional
    if t_eval.ndim > 1:
        raise ValueError("t_eval must be 1 dimensional.")

    out = jnp.append(jnp.append(t_span[0], t_eval), t_span[1])

    # output nan if t_eval point lies outside t_span
    t_min = jnp.min(t_span)
    t_max = jnp.max(t_span)
    out = cond(
        (jnp.min(t_eval) < t_min) | (jnp.max(t_eval) > t_max),
        lambda s: jnp.nan * s,
        lambda s: s,
        out,
    )

    # output nan if t_eval and t_span have incompatible orderings
    diff = jnp.diff(t_eval)
    t_direction = jnp.sign(t_span[1] - t_span[0])
    out = cond(jnp.any(t_direction * diff < 0.0), lambda s: jnp.nan * s, lambda s: s, out)

    # if out[0] == out[1], set out[1] == (out[2] + out[0])/2
    out = cond(out[0] == out[1], lambda x: x.at[1].set((x[2] + x[0]) / 2), lambda x: x, out)

    # if out[-1] == out[-2], set out[-2] == (out[-3] + out[-1])/2
    out = cond(out[-1] == out[-2], lambda x: x.at[-2].set((x[-3] + x[-1]) / 2), lambda x: x, out)

    return out


def trim_t_results_jax(
    results: OdeResult,
    t_eval: Optional[ArrayLike] = None,
) -> OdeResult:
    """JAX-compilable version of trim_t_results.

    Note the choice of which entry to remove in the case of duplicate time entries is due to
    peculiarities in :func:`jax_odeint`.

    Args:
        results: Result object, assumed to contain solution at time points from the output of
            ``merge_t_args_jax(t_span, t_eval)``.
        t_eval: Time points to include in returned results.

    Returns:
        OdeResult: Results with only times/solutions in ``t_eval``. If ``t_eval`` is ``None``,
            returns solver default output, with an additional correction for the possibility of
            ``t_span == [a, a]``.
    """

    if t_eval is not None:
        # remove second entry if t_eval[0] == results.t[0], as this indicates a repeated time
        results.y = cond(
            t_eval[0] == results.t[0],
            lambda y: jnp.append(jnp.array([y[0]]), y[2:], axis=0),
            lambda y: y[1:],
            jnp.array(results.y),
        )

        # remove second last entry if t_eval[-1] == results.t[-1], as this indicates a repeated time
        results.y = cond(
            t_eval[-1] == results.t[-1],
            lambda y: jnp.append(y[:-2], jnp.array([y[-1]]), axis=0),
            lambda y: y[:-1],
            jnp.array(results.y),
        )

        results.t = t_eval

    # this handles the odd case that t_span == [a, a]
    results.y = cond(
        results.t[0] == results.t[-1],
        lambda y: y.at[-1].set(y[0]),
        lambda y: y,
        jnp.array(results.y),
    )

    return results


def setup_args_lists(
    args_list: List[any], args_names: List[str], args_to_list: List[Callable]
) -> Tuple[List[List[any]], bool]:
    """Transform each entry of ``args_list`` into lists of the same length.

    All elements of args_list must be either given as lists of valid specifications,
    or as a singleton of a valid specification. ``args_to_list`` contains a list of functions
    that maps its corresponding entry in args_list to a valid list of singletons, along
    with a bool indicating whether the corresponding argument was already a list of singletons.
    All args are expanded to be the same length as the longest argument. For input in args_list:
        - If len(arg) == 1, it will be repeated max_len times
        - if len(arg) == max_len, nothing is done
        - If len(arg) not in (1, max_len), an error is raised

    Args:
        args_list: List of 'inputs'.
        args_names: Names of inputs, used for error raising.
        args_to_list: As described in the main function doc. Specialized error handling for
                      the entries in args_list should be part of these functions in the
                      case that the arg is neither a valid singleton, nor a valid list
                      of singletons.

    Returns:
        Tuple: First entry is the transformed version of args_list so that all entries are the
               lists of the same length. Second entry is a boolean stating whether the inputs
               were given all as single instances (False), or one was a list (True).

    Raises:
        QiskitError: If inputs have incompatible lengths.
    """

    args_as_lists = []
    args_were_lists = False
    for arg, to_list in zip(args_list, args_to_list):
        arg_as_list, arg_was_list = to_list(arg)
        args_as_lists.append(arg_as_list)
        args_were_lists = args_were_lists or arg_was_list

    arg_lens = [len(x) for x in args_as_lists]
    max_len = max(arg_lens)
    for idx, arg_len in enumerate(arg_lens):
        if arg_len not in (1, max_len):
            max_name = args_names[arg_lens.index(max_len)]

            arg_name_sequence = "".join(f"{name}, " for name in args_names[:-1])
            arg_name_sequence += f"and {args_names[-1]}"

            raise QiskitError(
                f"""If one of {arg_name_sequence} is given as a list of valid inputs, then the
                others must specify only a single input, or a list of the same length.
                {max_name} specifies {max_len} inputs, but {args_names[idx]} is of length {arg_len},
                which is incompatible."""
            )

    args_as_lists = [
        x * max_len if arg_len == 1 else x for x, arg_len in zip(args_as_lists, arg_lens)
    ]
    return args_as_lists, args_were_lists



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Perturbative solver base class.
"""

from abc import ABC, abstractmethod
from typing import Callable, List, Union, Optional

import numpy as np
from scipy.integrate._ivp.ivp import OdeResult

from qiskit import QiskitError

from qiskit_dynamics import DYNAMICS_NUMPY_ALIAS as numpy_alias
from qiskit_dynamics import Signal, ArrayLike
from qiskit_dynamics.arraylias.alias import _preferred_lib
from qiskit_dynamics.signals import SignalList

from .expansion_model import ExpansionModel
from ..solver_utils import setup_args_lists

try:
    import jax.numpy as jnp
    from jax import vmap
    from jax.lax import associative_scan
except ImportError:
    pass


class _PerturbativeSolver(ABC):
    """Abstract base class for perturbation-based solvers."""

    @abstractmethod
    def __init__(self, model: ExpansionModel):
        r"""Initialize.

        Args:
            model: ExpansionModel defining solver behaviour.
        """
        self._model = model

    @property
    def model(self) -> ExpansionModel:
        """Model object storing expansion details."""
        return self._model

    def solve(
        self,
        t0: Union[float, List[float]],
        n_steps: Union[int, List[int]],
        y0: Union[ArrayLike, List[ArrayLike]],
        signals: Union[List[Signal], List[List[Signal]]],
        jax_control_flow: Optional[bool] = None,
    ) -> Union[OdeResult, List[OdeResult]]:
        """Solve given an initial time, number of steps, signals, and initial state.

        Note that this method can be used to solve a list of simulations at once, by specifying
        one or more of the arguments ``t0``, ``n_steps``, ``y0``, or ``signals`` as a list of
        valid inputs. For this mode of operation, all of these arguments must be either lists of
        the same length, or a single valid input, which will be used repeatedly.

        Args:
            t0: Initial time.
            n_steps: Number of time steps to solve for.
            y0: Initial state at time ``t0``.
            signals: List of signals.
            jax_control_flow: Whether or not to use JAX control flow during solver runs. If
                unspecified it will be automatically determined based on the input state type,
                whether the call is taking place within a JAX transformation, or if the underlying
                expansion polynomial consists of JAX arrays.

        Returns:
            OdeResult: Results object, or list of results objects.

        Raises:
            QiskitError: If improperly formatted arguments.
        """

        if jax_control_flow is None:
            try:
                from jax import core

                jax_control_flow = (
                    ("jax" in numpy_alias.infer_libs(y0))
                    or isinstance(jnp.array(0), core.Tracer)
                    or (
                        "jax"
                        == _preferred_lib(
                            self.model.expansion_polynomial.constant_term,
                            self.model.expansion_polynomial.array_coefficients,
                        )
                    )
                )
            except ImportError:
                jax_control_flow = False

        # validate and setup list of simulations
        [t0_list, n_steps_list, y0_list, signals_list], multiple_sims = setup_args_lists(
            args_list=[t0, n_steps, y0, signals],
            args_names=["t0", "n_steps", "y0", "signals"],
            args_to_list=[
                lambda x: _scalar_to_list(x, "t0"),
                lambda x: _scalar_to_list(x, "n_steps"),
                _y0_to_list,
                _signals_to_list,
            ],
        )

        all_results = []
        for args in zip(t0_list, n_steps_list, y0_list, signals_list):
            if len(args[-1]) != len(self.model.operators):
                raise QiskitError("Signals must be the same length as the operators in the model.")
            all_results.append(
                self._solve(
                    t0=args[0],
                    n_steps=args[1],
                    y0=args[2],
                    signals=args[3],
                    jax_control_flow=jax_control_flow,
                )
            )

        if multiple_sims is False:
            return all_results[0]

        return all_results

    @abstractmethod
    def _solve(
        self,
        t0: float,
        n_steps: int,
        y0: ArrayLike,
        signals: List[Signal],
        jax_control_flow: bool = False,
    ) -> OdeResult:
        """Solve once for a list of signals, initial state, initial time, and number of steps.

        Args:
            t0: Initial time.
            n_steps: Number of time steps to solve for.
            y0: Initial state at time t0.
            signals: List of signals.
            jax_control_flow: Whether or not to use JAX control flow during solver runs.

        Returns:
            OdeResult: Results object.
        """
        pass


def _perturbative_solve(
    single_step: Callable,
    model: "ExpansionModel",
    signals: List[Signal],
    y0: np.ndarray,
    t0: float,
    n_steps: int,
) -> np.ndarray:
    """Standard python logic version of perturbative solving routine."""
    U0 = model.rotating_frame.state_out_of_frame(t0, np.eye(model.Udt.shape[0], dtype=complex))
    Uf = model.rotating_frame.state_into_frame(
        t0 + n_steps * model.dt, np.eye(U0.shape[0], dtype=complex)
    )

    sig_cheb_coeffs = model.approximate_signals(signals, t0, n_steps)

    y = U0 @ y0
    for k in range(n_steps):
        y = single_step(sig_cheb_coeffs[:, k], y)

    return Uf @ y


def _perturbative_solve_jax(
    single_step: Callable,
    model: "ExpansionModel",
    signals: List[Signal],
    y0: np.ndarray,
    t0: float,
    n_steps: int,
) -> np.ndarray:
    """JAX version of _perturbative_solve."""
    U0 = model.rotating_frame.state_out_of_frame(t0, jnp.eye(model.Udt.shape[0], dtype=complex))
    Uf = model.rotating_frame.state_into_frame(
        t0 + n_steps * model.dt, jnp.eye(U0.shape[0], dtype=complex)
    )

    sig_cheb_coeffs = model.approximate_signals(signals, t0, n_steps)

    y = U0 @ y0

    step_propagators = vmap(single_step)(jnp.flip(sig_cheb_coeffs.transpose(), axis=0))
    y = associative_scan(jnp.matmul, step_propagators, axis=0)[-1] @ y

    return Uf @ y


def _scalar_to_list(x, name):
    """Check if x is a scalar or a list of scalars, and convert to a list in either case."""
    was_list = False
    x_ndim = _nested_ndim(x)
    if x_ndim > 1:
        raise QiskitError(f"{name} must be either 0d or 1d.")

    if x_ndim == 1:
        was_list = True
    else:
        x = [x]

    return x, was_list


def _y0_to_list(y0):
    """Check if y0 is a single array or list of arrays, and return as a list in either case."""
    was_list = False
    if not isinstance(y0, list):
        y0 = [y0]
    else:
        was_list = True

    return y0, was_list


def _signals_to_list(signals):
    """Check if signals is a single signal specification or a list of
    such specifications, and return as a list in either case.
    """
    was_list = False
    if signals is None:
        signals = [signals]
    elif isinstance(signals, list) and isinstance(signals[0], (list, SignalList)):
        # multiple lists
        was_list = True
    elif isinstance(signals, SignalList) or (
        isinstance(signals, list) and not isinstance(signals[0], (list, SignalList))
    ):
        # single signals list
        signals = [signals]
    else:
        raise QiskitError("Signals specified in invalid format.")

    return signals, was_list


def _nested_ndim(x):
    """Determine the 'ndim' of x, which could be composed of nested lists and array types."""
    if isinstance(x, (list, tuple)):
        return 1 + _nested_ndim(x[0])
    elif hasattr(x, "ndim"):
        return x.ndim

    # assume scalar
    return 0
